# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Schemas for the plugin DOCUMENTATION data."""

import typing as t

import pydantic as p

from .base import BaseModel, DocSchema, OptionsSchema
from .plugin import PluginExamplesSchema, PluginMetadataSchema, PluginReturnSchema


class InnerModuleOptionsSchema(OptionsSchema):
    suboptions: t.Dict[str, 'InnerModuleOptionsSchema'] = {}

    @p.root_validator(pre=True)
    def allow_description_to_be_optional(cls, values):
        # Doing this in a validator so that the json-schema will still flag it as an error
        if 'description' not in values:
            values['description'] = []
        return values


InnerModuleOptionsSchema.update_forward_refs()


class ModuleOptionsSchema(OptionsSchema):
    suboptions: t.Dict[str, 'InnerModuleOptionsSchema'] = {}


class OuterModuleDocSchema(DocSchema):
    options: t.Dict[str, ModuleOptionsSchema] = {}


# Ignore Uninitialized attribute error as BaseModel works some magic to initialize the
# attributes when data is loaded into them.
# pyre-ignore[13]
class ModuleDocSchema(BaseModel):
    doc: OuterModuleDocSchema


class ModuleSchema(ModuleDocSchema, PluginExamplesSchema, PluginMetadataSchema,
                   PluginReturnSchema, BaseModel):
    """Documentation for modules."""



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Functions to deal with collections on the local system"""
import asyncio
import os
from typing import List

import sh

from .compat import best_get_loop


class CollectionFormatError(Exception):
    pass


async def install_together(collection_tarballs: List[str],
                           ansible_collections_dir: str) -> None:
    loop = best_get_loop()

    installers = []
    for pathname in collection_tarballs:
        namespace, collection, _dummy = os.path.basename(pathname).split('-', 2)
        collection_dir = os.path.join(ansible_collections_dir, namespace, collection)
        # Note: mkdir -p equivalent is okay because we created package_dir ourselves as a directory
        # that only we can access
        os.makedirs(collection_dir, mode=0o700, exist_ok=False)

        # If the choice of install tools for galaxy is ever settled upon, we can switch from tar to
        # using that
        # sh dynamically creates functions which map to executables
        # pyre-ignore[16]
        installers.append(loop.run_in_executor(None, sh.tar, '-xf', pathname, '-C', collection_dir))

    await asyncio.gather(*installers)


async def install_separately(collection_tarballs: List[str], collection_dir: str) -> List:
    loop = asyncio.get_running_loop()

    installers = []
    collection_dirs = []

    if not collection_tarballs:
        return collection_dirs

    for pathname in collection_tarballs:
        filename = os.path.basename(pathname)
        namespace, collection, version_ext = filename.split('-', 2)
        version = None
        for ext in ('.tar.gz',):
            # Note: If galaxy allows other archive formats, add their extensions here
            ext_start = version_ext.find(ext)
            if ext_start != -1:
                version = version_ext[:ext_start]
                break
        else:
            raise CollectionFormatError('Collection filename was in an unexpected'
                                        f' format: {filename}')

        package_dir = os.path.join(collection_dir, f'ansible-collections-{namespace}.'
                                   f'{collection}-{version}')
        os.mkdir(package_dir, mode=0o700)
        collection_dirs.append(package_dir)

        collection_dir = os.path.join(package_dir, 'ansible_collections', namespace, collection)
        # Note: this is okay because we created package_dir ourselves as a directory
        # that only we can access
        os.makedirs(collection_dir, mode=0o700, exist_ok=False)

        # If the choice of install tools for galaxy is ever settled upon, we can switch from tar to
        # using that
        # sh dynamically creates functions which map to executables
        # pyre-ignore[16]
        installers.append(loop.run_in_executor(None, sh.tar, '-xf', pathname,
                                               '-C', collection_dir))

    await asyncio.gather(*installers)

    return collection_dirs



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Entrypoint to the antsibull-docs script."""

import typing as t

from ...compat import asyncio_run
from ...docs_parsing.ansible_doc import get_ansible_plugin_info
from ...logging import log
from ...venv import FakeVenvRunner
from ...write_docs import output_indexes, output_all_plugin_rst
from .stable import normalize_all_plugin_info, get_collection_contents

if t.TYPE_CHECKING:
    import argparse


mlog = log.fields(mod=__name__)


def generate_docs(args: 'argparse.Namespace') -> int:
    """
    Create documentation for the current subcommand.

    Current documentation creates documentation for the currently installed version of Ansible,
    as well as the currently installed collections.

    :arg args: The parsed comand line args.
    :returns: A return code for the program.  See :func:`antsibull.cli.antsibull_docs.main` for
        details on what each code means.
    """
    flog = mlog.fields(func='generate_docs')
    flog.debug('Begin processing docs')

    venv = FakeVenvRunner()

    # Get the list of plugins
    plugin_info = asyncio_run(get_ansible_plugin_info(venv, args.collection_dir))
    flog.debug('Finished parsing info from plugins')

    """
    # Turn these into some sort of decorator that will load choose to dump or load the values
    # if a command line arg is specified.
    with open('dump_raw_plugin_info.json', 'w') as f:
        import json
        json.dump(plugin_info, f)
    flog.debug('Finished dumping raw plugin_info')
    """

    plugin_info, nonfatal_errors = asyncio_run(normalize_all_plugin_info(plugin_info))
    flog.debug('Finished normalizing data')
    # calculate_additional_info(plugin_info) (full_path)

    """
    with open('dump_normalized_plugin_info.json', 'w') as f:
        json.dump(plugin_info, f)
    flog.debug('Finished dumping normalized data')

    with open('dump_errors.json', 'w') as f:
        json.dump(nonfatal_errors, f)
    flog.debug('Finished dump errors')

    with open('dump_normalized_plugin_info.json', 'r') as f:
        import json
        plugin_info = json.load(f)
    flog.debug('Finished loading normalized data')

    with open('dump_errors.json', 'r') as f:
        from collections import defaultdict
        nonfatal_errors = json.load(f)
        nonfatal_errors = defaultdict(lambda: defaultdict(list), nonfatal_errors)
        for key, value in nonfatal_errors.items():
            nonfatal_errors[key] = defaultdict(list, value)
    flog.debug('Finished loading errors')
    """

    asyncio_run(output_all_plugin_rst(plugin_info, nonfatal_errors, args.dest_dir))
    flog.debug('Finished writing plugin docs')

    collection_info = get_collection_contents(plugin_info, nonfatal_errors)
    flog.debug('Finished writing collection data')

    asyncio_run(output_indexes(collection_info, args.dest_dir))
    flog.debug('Finished writing indexes')

    return 0



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Output documentation."""

import asyncio
import os.path
import typing as t

import aiofiles
import asyncio_pool

from jinja2 import Template

from .constants import THREAD_MAX
from .docs_parsing.fqcn import get_fqcn_parts
from .jinja2.environment import doc_environment

#: Mapping of plugins to nonfatal errors.  This is the type to use when accepting the plugin.
#: The mapping is of plugin_type: plugin_name: [error_msgs]
PluginErrorsT = t.Mapping[str, t.Mapping[str, t.Sequence[str]]]


async def write_rst(plugin_name: str, plugin_type: str, plugin_record: t.Dict[str, t.Any],
                    nonfatal_errors: PluginErrorsT, plugin_tmpl: Template, error_tmpl: Template,
                    dest_dir: str) -> None:
    """
    Write the rst page for one plugin.

    :arg plugin_name: FQCN for the plugin.
    :arg plugin_type: The type of the plugin.  (module, inventory, etc)
    :arg plugin_record: The record for the plugin.  doc, examples, and return are the
        toplevel fields.
    :arg nonfatal_errors: Mapping of plugin to any nonfatal errors that will be displayed in place
        of some or all of the docs
    :arg plugin_tmpl: Template for the plugin.
    :arg error_tmpl: Template to use when there wasn't enough documentation for the plugin.
    :arg dest_dir: Destination directory for the plugin data.  For instance,
        :file:`ansible-checkout/docs/docsite/rst/`.  The directory structure underneath this
        directory will be created if needed.
    """
    namespace, collection, plugin_short_name = get_fqcn_parts(plugin_name)
    collection_name = '.'.join((namespace, collection))

    if not plugin_record:
        plugin_contents = error_tmpl.render(
            plugin_type=plugin_type, plugin_name=plugin_name,
            collection=collection_name,
            nonfatal_errors=nonfatal_errors)
    else:
        plugin_contents = plugin_tmpl.render(
            collection=collection_name,
            plugin_type=plugin_type,
            plugin_name=plugin_name,
            doc=plugin_record['doc'],
            examples=plugin_record['examples'],
            returndocs=plugin_record['return_'],
            nonfatal_errors=nonfatal_errors)

    collection_dir = os.path.join(dest_dir, 'collections', namespace, collection)
    # This is dangerous but the code that takes dest_dir from the user checks
    # permissions on it to make it as safe as possible.
    os.makedirs(collection_dir, exist_ok=True)

    plugin_file = os.path.join(collection_dir, f'{plugin_short_name}_{plugin_type}.rst')

    async with aiofiles.open(plugin_file, 'w') as f:
        await f.write(plugin_contents)


async def output_all_plugin_rst(plugin_info: t.Dict[str, t.Any],
                                nonfatal_errors: PluginErrorsT,
                                dest_dir: str) -> None:
    """
    Output rst files for each plugin.

    :arg plugin_info: Documentation information for all of the plugins.
    :arg nonfatal_errors: Mapping of plugins to nonfatal errors.  Using this to note on the docs
        pages when documentation wasn't formatted such that we could use it.
    :arg dest_dir: The directory to place the documentation in.
    """
    # Setup the jinja environment
    env = doc_environment(('antsibull.data', 'docsite'))
    # Get the templates
    plugin_tmpl = env.get_template('plugin.rst.j2')
    error_tmpl = env.get_template('plugin-error.rst.j2')

    writers = []
    async with asyncio_pool.AioPool(size=THREAD_MAX) as pool:
        for plugin_type, plugins_by_type in plugin_info.items():
            for plugin_name, plugin_record in plugins_by_type.items():
                writers.append(pool.spawn_n(write_rst(plugin_name, plugin_type, plugin_record,
                                                      nonfatal_errors[plugin_type][plugin_name],
                                                      plugin_tmpl, error_tmpl, dest_dir)))

        # Write docs for each plugin
        await asyncio.gather(*writers)


async def write_collection_list(collections: t.Iterable[str], template: Template,
                                dest_dir: str) -> None:
    """
    Write an index page listing all of the collections.

    Each collection will link to an index page listing all content in the collection.

    :arg collections: Iterable of all the collection names.
    :arg template: A template to render the collection index.
    :arg dest_dir: The destination directory to output the index into.
    """
    index_contents = template.render(collections=collections)
    index_file = os.path.join(dest_dir, 'index.rst')

    async with aiofiles.open(index_file, 'w') as f:
        await f.write(index_contents)


async def write_plugin_lists(collection_name: str,
                             plugin_maps: t.Mapping[str, t.Mapping[str, str]],
                             template: Template,
                             dest_dir: str) -> None:
    """
    Write an index page for each collection.

    The per-collection index page links to plugins for each collection.

    :arg plugin_maps: Mapping of plugin_type to Mapping of plugin_name to short_description.
    :arg template: A template to render the collection index.
    :arg dest_dir: The destination directory to output the index into.
    """
    index_contents = template.render(
        collection_name=collection_name,
        plugin_maps=plugin_maps)

    index_file = os.path.join(dest_dir, 'index.rst')

    async with aiofiles.open(index_file, 'w') as f:
        await f.write(index_contents)


async def output_indexes(collection_info: t.Mapping[str, t.Mapping[str, t.Mapping[str, str]]],
                         dest_dir: str) -> None:
    """
    Generate index pages for the collections.

    :arg collection_info: Mapping of collection_name to Mapping of plugin_type to Mapping of
        collection_name to short_description.
    :arg dest_dir: The directory to place the documentation in.
    """
    env = doc_environment(('antsibull.data', 'docsite'))
    # Get the templates
    collection_list_tmpl = env.get_template('list_of_collections.rst.j2')
    collection_plugins_tmpl = env.get_template('plugins_by_collection.rst.j2')

    writers = []
    async with asyncio_pool.AioPool(size=THREAD_MAX) as pool:
        collection_toplevel = os.path.join(dest_dir, 'collections')
        writers.append(pool.spawn_n(write_collection_list(collection_info.keys(),
                                                          collection_list_tmpl,
                                                          collection_toplevel)))

        for collection_name, plugin_maps in collection_info.items():
            collection_dir = os.path.join(collection_toplevel, *(collection_name.split('.')))
            writers.append(pool.spawn_n(write_plugin_lists(collection_name,
                                                           plugin_maps,
                                                           collection_plugins_tmpl,
                                                           collection_dir)))

        await asyncio.gather(*writers)



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Functions for working with the ansible-base package."""
import ast
import os
import re
import typing as t
from functools import lru_cache
from urllib.parse import urljoin

import aiofiles
import packaging.version as pypiver
import sh

from .compat import best_get_loop
from .constants import CHUNKSIZE

if t.TYPE_CHECKING:
    import aiohttp.client


#: URL to checkout ansible-base from.
ANSIBLE_BASE_URL = 'https://github.com/ansible/ansible'
#: URL to pypi.
PYPI_SERVER_URL = 'https://test.pypi.org/'


class UnknownVersion(Exception):
    """Raised when a requested version does not exist."""


class AnsibleBasePyPiClient:
    """Class to retrieve information about AnsibleBase from Pypi."""

    def __init__(self, aio_session: 'aiohttp.client.ClientSession',
                 pypi_server_url: str = PYPI_SERVER_URL) -> None:
        """
        Initialize the AnsibleBasePypi class.

        :arg aio_session: :obj:`aiohttp.client.ClientSession` to make requests to pypi from.
        :kwarg pypi_server_url: URL to the pypi server to use.
        """
        self.aio_session = aio_session
        self.pypi_server_url = pypi_server_url

    @lru_cache(None)
    async def get_info(self) -> t.Dict[str, t.Any]:
        """
        Retrieve information about the ansible-base package from pypi.

        :returns: The dict which represents the information about the ansible-base package returned
            from pypi.  To examine the data structure, use::

                curl https://pypi.org/pypi/ansible-base/json| python3 -m json.tool
        """
        # Retrieve the ansible-base package info from pypi
        query_url = urljoin(self.pypi_server_url, 'pypi/ansible-base/json')
        async with self.aio_session.get(query_url) as response:
            pkg_info = await response.json()
        return pkg_info

    async def get_versions(self) -> t.List[pypiver.Version]:
        """
        Get the versions of the ansible-base package on pypi.

        :returns: A list of :pypkg:obj:`packaging.versioning.Version`s
            for all the versions on pypi, including prereleases.
        """
        pkg_info = await self.get_info()
        versions = [pypiver.Version(r) for r in pkg_info['releases']]
        versions.sort(reverse=True)
        return versions

    async def get_latest_version(self) -> pypiver.Version:
        """
        Get the latest version of ansible-base uploaded to pypi.

        :return: A :pypkg:obj:`packaging.versioning.Version` object representing the latest version
            of the package on pypi.  This may be a pre-release.
        """
        versions = await self.get_versions()
        return versions[0]

    async def retrieve(self, ansible_base_version: str, download_dir: str) -> str:
        """
        Get the release from pypi.

        :arg ansible_base_version: Version of ansible-base to download.
        :arg download_dir: Directory to download the tarball to.
        :returns: The name of the downloaded tarball.
        """
        pkg_info = await self.get_info()

        pypi_url = tar_filename = ''
        for release in pkg_info['releases'][ansible_base_version]:
            if release['filename'].startswith(f'ansible-base-{ansible_base_version}.tar.'):
                tar_filename = release['filename']
                pypi_url = release['url']
                break
        else:  # for-else: http://bit.ly/1ElPkyg
            raise UnknownVersion(f'ansible-base {ansible_base_version} does not'
                                 ' exist on {pypi_server_url}')

        tar_filename = os.path.join(download_dir, tar_filename)
        async with self.aio_session.get(pypi_url) as response:
            async with aiofiles.open(tar_filename, 'wb') as f:
                # TODO: PY3.8: while chunk := await response.read(CHUNKSIZE):
                chunk = await response.content.read(CHUNKSIZE)
                while chunk:
                    await f.write(chunk)
                    chunk = await response.content.read(CHUNKSIZE)

        return tar_filename


def _get_cache_version(ansible_base_cache: str) -> pypiver.Version:
    with open(os.path.join(ansible_base_cache, 'lib', 'ansible', 'release.py')) as f:
        root = ast.parse(f.read())

    # Find the version of the cache
    cache_version = None
    # Iterate backwards in case __version__ is assigned to multiple times
    for node in reversed(root.body):
        if isinstance(node, ast.Assign):
            for name in node.targets:
                # These attributes are dynamic so pyre cannot check them
                if name.id == '__version__':  # pyre-ignore[16]
                    cache_version = node.value.s  # pyre-ignore[16]
                    break

        if cache_version:
            break

    if not cache_version:
        raise ValueError('Version was not found')

    return pypiver.Version(cache_version)


def cache_is_devel(ansible_base_cache: t.Optional[str]) -> bool:
    """
    :arg ansible_base_cache: A path to an Ansible-base checkout or expanded sdist or None.
        This will be used instead of downloading an ansible-base package if the version matches
        with ``ansible_base_version``.
    :returns: True if the cache looks like it is for the devel branch.
    """
    if ansible_base_cache is None:
        return False

    try:
        cache_version = _get_cache_version(ansible_base_cache)
    except Exception:
        return False

    dev_version = re.compile('[.]dev[0-9]+$')
    if dev_version.match(cache_version.public):
        return True

    return False


def cache_is_correct_version(ansible_base_cache: t.Optional[str],
                             ansible_base_version: pypiver.Version) -> bool:
    """
    :arg ansible_base_cache: A path to an Ansible-base checkout or expanded sdist or None.
        This will be used instead of downloading an ansible-base package if the version matches
        with ``ansible_base_version``.
    :arg ansible_base_version: Version of ansible-base to retrieve.
    :returns: True if the cache is for a compatible version at or newer than the requested version
    """
    if ansible_base_cache is None:
        return False

    try:
        cache_version = _get_cache_version(ansible_base_cache)
    except Exception:
        return False

    # If the cache is a compatible version of ansible-base and it is the same or more recent than
    # the requested version then allow this.
    if (cache_version.major == ansible_base_version.major
            and cache_version.minor == ansible_base_version.minor
            and cache_version.micro >= ansible_base_version.micro):
        return True

    return False


@lru_cache(None)
async def checkout_from_git(download_dir: str, repo_url: str = ANSIBLE_BASE_URL) -> str:
    """
    Checkout the ansible-base git repo.

    :arg download_dir: Directory to checkout into.
    :kwarg: repo_url: The url to the git repo.
    :return: The directory that ansible-base has been checked out to.
    """
    loop = best_get_loop()
    ansible_base_dir = os.path.join(download_dir, 'ansible-base')
    await loop.run_in_executor(None, sh.git, 'clone', repo_url, ansible_base_dir)

    return ansible_base_dir


async def get_ansible_base(aio_session: 'aiohttp.client.ClientSession',
                           ansible_base_version: str,
                           tmpdir: str,
                           ansible_base_cache: t.Optional[str] = None) -> str:
    """
    Create an ansible-base directory of the requested version.

    :arg aio_session: :obj:`aiohttp.client.ClientSession` to make http requests with.
    :arg ansible_base_version: Version of ansible-base to retrieve.
    :arg tmpdir: Temporary directory use as a scratch area for downloading to and the place that the
        ansible-base directory should be placed in.
    :kwarg ansible_base_cache: If given, a path to an Ansible-base checkout or expanded sdist.
        This will be used instead of downloading an ansible-base package if the version matches
        with ``ansible_base_version``.
    """
    if ansible_base_version == '@devel':
        # is the cache usable?
        if cache_is_devel(ansible_base_cache):
            assert ansible_base_cache is not None
            return ansible_base_cache

        install_file = await checkout_from_git(tmpdir)
    else:
        pypi_client = AnsibleBasePyPiClient(aio_session)
        if ansible_base_version == '@latest':
            ansible_base_version: pypiver.Version = await pypi_client.get_latest_version()
        else:
            ansible_base_version: pypiver.Version = pypiver.Version(ansible_base_version)

        # is the cache the asked for version?
        if cache_is_correct_version(ansible_base_cache, ansible_base_version):
            assert ansible_base_cache is not None
            return ansible_base_cache

        install_file = await pypi_client.retrieve(ansible_base_version.public, tmpdir)

    return install_file



# Copyright: (c) 2019, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

import warnings
from distutils.version import LooseVersion
from functools import partial

from ..vendored.collections import is_sequence


# if a module is added in a version of Ansible older than this, don't print the version added
# information in the module documentation because everyone is assumed to be running something newer
# than this already.
TOO_OLD_TO_BE_NOTABLE = '0'

test_list = partial(is_sequence, include_strings=False)


def still_relevant(version, cutoff=TOO_OLD_TO_BE_NOTABLE):
    """
    Calculates whether the given version is older than a cutoff value

    :arg version: Version to check
    :arg cutoff: Calculate whether `version` is older than this
    :returns: True if the `version` is older than `cutoff` otherwise True.

    .. note:: This is similar to the ansible `version_compare` test but needs to handle the
        `historical` version and empty version.
    """
    # Note: This was the opposite in previous code but then the version_added was stripped out by
    # other things
    if not version:
        return False

    if version == 'historical':
        return False

    # In case it was specified as a string or float in yaml
    try:
        version = LooseVersion(version)
    except ValueError as e:
        warnings.warn("Could not parse %s: %s" % (version, str(e)))
        return True
    try:
        return version >= LooseVersion(cutoff)
    except Exception as e:
        warnings.warn("Could not compare %s: %s" % (version, str(e)))
        return True



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""
Schemas for the plugin DOCUMENTATION data.

This is a highlevel interface.  The hope is that developers can use either this or
antsibull.schemas.docs to handle all of their validation needs.
"""

# Ignore Unitialized attribute errors because BaseModel works some magic
# to initialize the attributes when data is loaded into them.
# pyre-ignore-all-errors[13]

import typing as t

from .base import BaseModel
from .callback import CallbackSchema
from .module import ModuleSchema
from .plugin import PluginSchema


class AnsibleDocSchema(BaseModel):
    """
    Document the output that comes back from ansible-doc.

    This schema is a constructed schema.  It imagines that the user has accumulated the output from
    repeated runs of ansible-doc in a dict layed out roughly like this:

    .. code-block:: yaml

        plugin_type:
            plugin_name:
                # json.loads(ansible-doc -t plugin_type --json plugin_name)

    Pass the dict above to :meth:`AnsibleDocSchema.parse_obj` to build the models from the schema.

    If you want to use the Schema to validate and normalize the data but need a :python:obj:`dict`
    afterwards, call :meth:`AnsibleDocSchema.dict` on the populated model to get
    a :python:obj:`dict` back out.
    """

    become: t.Dict[str, PluginSchema]
    cache: t.Dict[str, PluginSchema]
    callback: t.Dict[str, CallbackSchema]
    cliconf: t.Dict[str, PluginSchema]
    connection: t.Dict[str, PluginSchema]
    httpapi: t.Dict[str, PluginSchema]
    inventory: t.Dict[str, PluginSchema]
    lookup: t.Dict[str, PluginSchema]
    module: t.Dict[str, ModuleSchema]
    netconf: t.Dict[str, PluginSchema]
    shell: t.Dict[str, PluginSchema]
    strategy: t.Dict[str, PluginSchema]
    vars: t.Dict[str, PluginSchema]


class GenericPluginSchema(BaseModel):
    """
    Document the output of ``ansible-doc -t PLUGIN_TYPE PLUGIN_NAME``.

    .. note:: Both the model and the dict will be wrapped in an outer dict with your data mapped
        to the ``__root__`` key. This happens because the toplevel key of ansible-doc's output is
        a dynamic key which we can't automatically map to an attribute name.
    """

    __root__: t.Dict[str, PluginSchema]


class CallbackPluginSchema(BaseModel):
    """
    Document the output of ``ansible-doc -t callback CALLBACK_NAME``.

    .. note:: Both the model and the dict will be wrapped in an outer dict with your data mapped
        to the ``__root__`` key. This happens because the toplevel key of ansible-doc's output is
        a dynamic key which we can't automatically map to an attribute name.
    """

    __root__: t.Dict[str, CallbackSchema]


class ModulePluginSchema(BaseModel):
    """
    Document the output of ``ansible-doc -t module MODULE_NAME``.

    .. note:: Both the model and the dict will be wrapped in an outer dict with your data mapped
        to the ``__root__`` key. This happens because the toplevel key of ansible-doc's output is
        a dynamic key which we can't automatically map to an attribute name.
    """

    __root__: t.Dict[str, ModuleSchema]


#: Make sure users can access plugins using the plugin type rather than having to guess that
#: these types use the GenericPluginSchema
BecomePluginSchema = GenericPluginSchema
CachePluginSchema = GenericPluginSchema
CliConfPluginSchema = GenericPluginSchema
ConnectionPluginSchema = GenericPluginSchema
HttpApiPluginSchema = GenericPluginSchema
InventoryPluginSchema = GenericPluginSchema
LookupPluginSchema = GenericPluginSchema
NetConfPluginSchema = GenericPluginSchema
ShellPluginSchema = GenericPluginSchema
StrategyPluginSchema = GenericPluginSchema
VarsPluginSchema = GenericPluginSchema


#: A mapping from plugin type to the Schema to use for them.  Use this to more easily get
#: the Schema programmatically.
ANSIBLE_DOC_SCHEMAS = {
    'become': BecomePluginSchema,
    'cache': CachePluginSchema,
    'callback': CallbackPluginSchema,
    'cliconf': CliConfPluginSchema,
    'connection': ConnectionPluginSchema,
    'httpapi': HttpApiPluginSchema,
    'inventory': InventoryPluginSchema,
    'lookup': LookupPluginSchema,
    'module': ModulePluginSchema,
    'netconf': NetConfPluginSchema,
    'shell': ShellPluginSchema,
    'strategy': StrategyPluginSchema,
    'vars': VarsPluginSchema,
}



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Functions to help with hashing."""

import hashlib

import aiofiles

from .constants import CHUNKSIZE


async def verify_hash(filename: str, hash: str, algorithm: str = 'sha256') -> bool:
    """
    Verify whether a file has a given sha256sum.

    :arg filename: The file to verify the sha256sum of.
    :arg hash: The hash that is expected.
    :kwarg algorithm: The hash algorithm to use.  This must be present in hashlib on this
        system.  The default is 'sha256'
    :returns: True if the hash matches, otherwise False.
    """
    hasher = getattr(hashlib, algorithm)()
    async with aiofiles.open(filename, 'rb') as f:
        # TODO: PY3.8: while chunk := await f.read(CHUNKSIZE):
        chunk = await f.read(CHUNKSIZE)
        while chunk:
            hasher.update(chunk)
            chunk = await f.read(CHUNKSIZE)
    if hasher.hexdigest() != hash:
        return False

    return True



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Entrypoint to the antsibull-docs script."""

import asyncio
import os
import os.path
import tempfile
import typing as t
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor

import aiohttp
from pydantic import ValidationError

from ...ansible_base import get_ansible_base
from ...collections import install_together
from ...compat import asyncio_run, best_get_loop
from ...dependency_files import DepsFile
from ...docs_parsing.ansible_doc import get_ansible_plugin_info
from ...docs_parsing.fqcn import get_fqcn_parts
from ...galaxy import CollectionDownloader
from ...logging import log
from ...schemas.docs import DOCS_SCHEMAS
from ...venv import VenvRunner
from ...write_docs import output_indexes, output_all_plugin_rst

if t.TYPE_CHECKING:
    import argparse
    import semantic_version as semver


mlog = log.fields(mod=__name__)

#: Mapping of plugins to nonfatal errors.  This is the type to use when returning the mapping.
PluginErrorsRT = t.DefaultDict[str, t.DefaultDict[str, t.List[str]]]


async def retrieve(ansible_base_version: str,
                   collections: t.Mapping[str, str],
                   tmp_dir: str,
                   ansible_base_cache: t.Optional[str] = None,
                   collection_cache: t.Optional[str] = None) -> t.Dict[str, 'semver.Version']:
    """
    Download ansible-base and the collections.

    :arg ansible_base_version: Version of ansible-base to download.
    :arg collections: Map of collection names to collection versions to download.
    :arg tmp_dir: The directory to download into
    :kwarg ansible_base_cache: If given, a path to an Ansible-base checkout or expanded sdist.
        This will be used instead of downloading an ansible-base package if the version matches
        with ``ansible_base_version``.
    :kwarg collection_cache: If given, a path to a directory containing collection tarballs.
        These tarballs will be used instead of downloading new tarballs provided that the
        versions match the criteria (latest compatible version known to galaxy).
    :returns: Map of collection name to directory it is in.  ansible-base will
        use the special key, `_ansible_base`.
    """
    collection_dir = os.path.join(tmp_dir, 'collections')
    os.mkdir(collection_dir, mode=0o700)

    requestors = {}
    async with aiohttp.ClientSession() as aio_session:
        requestors['_ansible_base'] = asyncio.create_task(
            get_ansible_base(aio_session, ansible_base_version, tmp_dir,
                             ansible_base_cache=ansible_base_cache))

        downloader = CollectionDownloader(aio_session, collection_dir,
                                          collection_cache=collection_cache)
        for collection, version in collections.items():
            requestors[collection] = asyncio.create_task(downloader.download(collection, version))

        responses = await asyncio.gather(*requestors.values())

    # Note: Python dicts have always had a stable order as long as you don't modify the dict.
    # So requestors (implicitly, the keys) and responses have a matching order here.
    return dict(zip(requestors, responses))


def normalize_plugin_info(plugin_type: str,
                          plugin_info: t.Mapping[str, t.Any]
                          ) -> t.Tuple[t.Dict[str, t.Any], t.List[str]]:
    """
    Normalize and validate all of the plugin docs.

    :arg plugin_type: The type of plugins that we're getting docs for.
    :arg plugin_info: Mapping of plugin_info.  The toplevel keys are plugin names.
        See the schema in :mod:`antsibull.schemas` for what the data should look like and just how
        much conversion we can perform on it.
    :returns: A tuple containing a "copy" of plugin_info with all of the data normalized and a list
        of nonfatal errors.  The plugin_info dict will follow the structure expressed in the schemas
        in :mod:`antsibull.schemas`.  The nonfatal errors are strings representing the problems
        encountered.
    """
    new_info = {}
    errors = []
    # Note: loop through "doc" before any other keys.
    for field in ('doc', 'examples', 'return'):
        try:
            new_info.update(DOCS_SCHEMAS[plugin_type][field](**{field: plugin_info[field]}).dict())
        except ValidationError as e:
            if field == 'doc':
                # We can't recover if there's not a doc field
                raise
            # But we can use the default value (some variant of "empty") for everything else
            # Note: We looped through doc first and raised an exception if doc did not normalize
            # so we're able to use it in the error message here.
            errors.append(f'Unable to normalize {new_info["doc"]["name"]}: {field}'
                          f' due to: {str(e)}')
            new_info.update(DOCS_SCHEMAS[plugin_type][field].parse_obj({}))

    return (new_info, errors)


async def normalize_all_plugin_info(plugin_info: t.Mapping[str, t.Mapping[str, t.Any]]
                                    ) -> t.Tuple[t.Dict[str, t.Dict[str, t.Any]], PluginErrorsRT]:
    """
    Normalize the data in plugin_info so that it is ready to be passed to the templates.

    :arg plugin_info: Mapping of information about plugins.  This contains information about all of
        the plugins that are to be documented. See the schema in :mod:`antsibull.schemas` for the
        structure of the information.
    :returns: A tuple of plugin_info (this is a "copy" of the input plugin_info with all of the
        data normalized) and a mapping of errors.  The plugin_info may have less records than the
        input plugin_info if there were plugin records which failed to validate.  The mapping of
        errors takes the form of:

        .. code-block:: yaml

            plugin_type:
                plugin_name:
                    - error string
                    - error string
    """
    loop = best_get_loop()

    # Normalize each plugin in a subprocess since normalization is CPU bound
    normalizers = {}
    for plugin_type, plugin_list_for_type in plugin_info.items():
        for plugin_name, plugin_record in plugin_list_for_type.items():
            normalizers[(plugin_type, plugin_name)] = loop.run_in_executor(
                ProcessPoolExecutor(), normalize_plugin_info, plugin_type, plugin_record)

    results = await asyncio.gather(*normalizers.values(), return_exceptions=True)

    new_plugin_info = defaultdict(dict)
    nonfatal_errors = defaultdict(lambda: defaultdict(list))
    for (plugin_type, plugin_name), plugin_record in zip(normalizers, results):
        # Errors which broke doc parsing (and therefore we won't have enough info to
        # build a docs page)
        if isinstance(plugin_record, Exception):
            # An exception means there is no usable documentation for this plugin
            # Record a nonfatal error and then move on
            nonfatal_errors[plugin_type][plugin_name].append(str(plugin_record))
            continue

        # Errors where we have at least docs.  We can still create a docs page for these with some
        # information left out
        if plugin_record[1]:
            nonfatal_errors[plugin_type][plugin_name].extend(plugin_record[1])

        new_plugin_info[plugin_type][plugin_name] = plugin_record[0]

    return new_plugin_info, nonfatal_errors


def get_collection_contents(plugin_info: t.Mapping[str, t.Mapping[str, t.Any]],
                            nonfatal_errors: PluginErrorsRT
                            ) -> t.DefaultDict[str, t.DefaultDict[str, t.Dict[str, str]]]:
    """
    Return the contents plugins which are in each collection.

    :arg plugin_info: Mapping of plugin type to a mapping of plugin name to plugin record.
        The plugin_type, plugin_name, and short_description from plugin_records are used.
    :arg nonfatal_errors: mapping of plugin type to plugin name to list of error messages.
        The plugin_type and plugin_name are used.
    :returns: A Mapping of collection name to a mapping of plugin type to a mapping of plugin names
        to short_descriptions.
    collection:
        plugin_type:
            - plugin_short_name: short_description
    """
    collection_plugins = defaultdict(lambda: defaultdict(dict))
    # Some plugins won't have an entry in the plugin_info because documentation failed to parse.
    # Those should be documented in the nonfatal_errors information.
    for plugin_type, plugin_list in nonfatal_errors.items():
        for plugin_name, dummy_ in plugin_list.items():
            namespace, collection, short_name = get_fqcn_parts(plugin_name)
            collection_plugins['.'.join((namespace, collection))][plugin_type][short_name] = ''

    for plugin_type, plugin_list in plugin_info.items():
        for plugin_name, plugin_desc in plugin_list.items():
            namespace, collection, short_name = get_fqcn_parts(plugin_name)
            collection_plugins['.'.join((namespace, collection))][plugin_type][short_name] = (
                plugin_desc['doc']['short_description'])

    return collection_plugins


def generate_docs(args: 'argparse.Namespace') -> int:
    """
    Create documentation for the stable subcommand.

    Stable documentation creates documentation for a built version of Ansible.  It uses the exact
    versions of collections included in the last Ansible release to generate rst files documenting
    those collections.

    :arg args: The parsed comand line args.
    :returns: A return code for the program.  See :func:`antsibull.cli.antsibull_docs.main` for
        details on what each code means.
    """
    flog = mlog.fields(func='generate_docs')
    flog.debug('Begin processing docs')

    # Parse the deps file
    deps_file = DepsFile(args.deps_file)
    dummy_, ansible_base_version, collections = deps_file.parse()
    flog.debug('Finished parsing deps file')

    with tempfile.TemporaryDirectory() as tmp_dir:
        # Retrieve ansible-base and the collections
        collection_tarballs = asyncio_run(retrieve(ansible_base_version, collections, tmp_dir,
                                                   ansible_base_cache=args.ansible_base_cache,
                                                   collection_cache=args.collection_cache))
        flog.debug('Finished retrieving tarballs')

        # Get the ansible-base location
        try:
            # Note, this may be a tarball or the path to an ansible-base checkout/expanded sdist.
            ansible_base_path = collection_tarballs.pop('_ansible_base')
        except KeyError:
            print('ansible-base did not download successfully')
            return 3

        # Install the collections to a directory

        # Directory that ansible needs to see
        collection_dir = os.path.join(tmp_dir, 'installed')
        # Directory that the collections will be untarred inside of
        collection_install_dir = os.path.join(collection_dir, 'ansible_collections')
        # Safe to recursively mkdir because we created the tmp_dir
        os.makedirs(collection_install_dir, mode=0o700)

        # Install the collections
        asyncio_run(install_together(collection_tarballs.values(), collection_install_dir))
        flog.debug('Finished installing collections')

        # Create venv for ansible-base
        venv = VenvRunner('ansible-base-venv', tmp_dir)
        if os.path.isdir(ansible_base_path):
            venv.install_package(ansible_base_path, from_project_path=True)
        else:
            venv.install_package(ansible_base_path)
        flog.debug('Finished installing ansible-base')

        # Get the list of plugins
        plugin_info = asyncio_run(get_ansible_plugin_info(venv, collection_dir))
        flog.debug('Finished parsing info from plugins')

        """
        # Turn these into some sort of decorator that will load choose to dump or load the values
        # if a command line arg is specified.
        with open('dump_raw_plugin_info.json', 'w') as f:
            import json
            json.dump(plugin_info, f)
        flog.debug('Finished dumping raw plugin_info')
        """

        plugin_info, nonfatal_errors = asyncio_run(normalize_all_plugin_info(plugin_info))
        flog.debug('Finished normalizing data')
        # calculate_additional_info(plugin_info) (full_path)

        """
        with open('dump_normalized_plugin_info.json', 'w') as f:
            json.dump(plugin_info, f)
        flog.debug('Finished dumping normalized data')

        with open('dump_errors.json', 'w') as f:
            json.dump(nonfatal_errors, f)
        flog.debug('Finished dump errors')

        with open('dump_normalized_plugin_info.json', 'r') as f:
            import json
            plugin_info = json.load(f)
        flog.debug('Finished loading normalized data')

        with open('dump_errors.json', 'r') as f:
            from collections import defaultdict
            nonfatal_errors = json.load(f)
            nonfatal_errors = defaultdict(lambda: defaultdict(list), nonfatal_errors)
            for key, value in nonfatal_errors.items():
                nonfatal_errors[key] = defaultdict(list, value)
        flog.debug('Finished loading errors')
        """

        asyncio_run(output_all_plugin_rst(plugin_info, nonfatal_errors, args.dest_dir))
        flog.debug('Finished writing plugin docs')

        collection_info = get_collection_contents(plugin_info, nonfatal_errors)
        flog.debug('Finished writing collection data')

        asyncio_run(output_indexes(collection_info, args.dest_dir))
        flog.debug('Finished writing indexes')

    return 0



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020


def generate_docs(args) -> int:
    raise NotImplementedError('Priority to implement subcommands is stable, devel, plugin, and'
                              ' then collection commands')



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Parse documentation from ansible plugins using anible-doc."""

import asyncio
import json
import os
import sys
import traceback
from typing import TYPE_CHECKING, Any, Dict, Union

import sh

from .fqcn import get_fqcn_parts
from ..compat import best_get_loop
from ..constants import DOCUMENTABLE_PLUGINS
from ..vendored.json_utils import _filter_non_json_lines

if TYPE_CHECKING:
    from ..venv import VenvRunner, FakeVenvRunner


#: Clear Ansible environment variables that set paths where plugins could be found.
ANSIBLE_PATH_ENVVARS: Dict[str, str] = {'ANSIBLE_COLLECTIONS_PATHS': "/dev/null",
                                        'ANSIBLE_ACTION_PLUGINS': "/dev/null",
                                        'ANSIBLE_CACHE_PLUGINS': "/dev/null",
                                        'ANSIBLE_CALLBACK_PLUGINS': "/dev/null",
                                        'ANSIBLE_CLICONF_PLUGINS': "/dev/null",
                                        'ANSIBLE_CONNECTION_PLUGINS': "/dev/null",
                                        'ANSIBLE_FILTER_PLUGINS': "/dev/null",
                                        'ANSIBLE_HTTPAPI_PLUGINS': "/dev/null",
                                        'ANSIBLE_INVENTORY_PLUGINS': "/dev/null",
                                        'ANSIBLE_LOOKUP_PLUGINS': "/dev/null",
                                        'ANSIBLE_LIBRARY': "/dev/null",
                                        'ANSIBLE_MODULE_UTILS': "/dev/null",
                                        'ANSIBLE_NETCONF_PLUGINS': "/dev/null",
                                        'ANSIBLE_ROLES_PATH': "/dev/null",
                                        'ANSIBLE_STRATEGY_PLUGINS': "/dev/null",
                                        'ANSIBLE_TERMINAL_PLUGINS': "/dev/null",
                                        'ANSIBLE_TEST_PLUGINS': "/dev/null",
                                        'ANSIBLE_VARS_PLUGINS': "/dev/null",
                                        'ANSIBLE_DOC_FRAGMENT_PLUGINS': "/dev/null",
                                        }


class ParsingError(Exception):
    """Error raised while parsing plugins for documentation."""


async def _get_plugin_info(plugin_type: str, ansible_doc: 'sh.Command') -> Dict[str, Any]:
    """
    Retrieve info about all Ansible plugins of a particular type.

    :arg plugin_type: The type of plugin.  See :attr:`DOCUMENTABLE_PLUGINS` for a list
        of allowed types.
    :arg ansible_doc: An :sh:obj:`sh.Command` object that will run the ansible-doc command.
        This command should already have been baked with any necessary environment and
        common arguments.
        :returns: A nested dictionary structure that looks like::
            Dictionary mapping to plugin information.  The dictionary looks like::

                plugin_name:  # This is the canonical name for the plugin and includes
                              # namespace.collection_name unless the plugin comes from
                              # ansible-base.
                    {information from ansible-doc --json.  See the ansible-doc documentation for
                     more info.}
    """
    # Get the list of plugins
    ansible_doc_list_cmd = ansible_doc('--list', '--t', plugin_type, '--json')
    raw_plugin_list = ansible_doc_list_cmd.stdout.decode('utf-8', errors='surrogateescape')
    # Note: Keep ansible_doc_list_cmd around until we know if we need to use it in an error message.
    plugin_map = json.loads(_filter_non_json_lines(raw_plugin_list)[0])
    del raw_plugin_list
    del ansible_doc_list_cmd

    loop = best_get_loop()

    # For each plugin, get its documentation
    extractors = {}
    for plugin_name in plugin_map.keys():
        extractors[plugin_name] = loop.run_in_executor(None, ansible_doc, '-t', plugin_type,
                                                       '--json', plugin_name)
    plugin_info = await asyncio.gather(*extractors.values(), return_exceptions=True)

    results = {}
    for plugin_name, ansible_doc_results in zip(extractors, plugin_info):
        err_msg = []
        stdout = ansible_doc_results.stdout.decode("utf-8", errors="surrogateescape")
        stderr = ansible_doc_results.stderr.decode("utf-8", errors="surrogateescape")

        if isinstance(ansible_doc_results, Exception):
            formatted_exception = traceback.format_exception(None, ansible_doc_results,
                                                             ansible_doc_results.__traceback__)
            err_msg.append(f'Exception while parsing documentation for {plugin_type} plugin:'
                           f' {plugin_name}.  Will not document this plugin.')
            err_msg.append(f'Exception:\n{"".join(formatted_exception)}')

        # Note: Exception will also be True.
        if isinstance(ansible_doc_results, sh.ErrorReturnCode):
            err_msg.append(f'Full process stdout:\n{stdout}')
            err_msg.append(f'Full process stderr:\n{stderr}')

        if err_msg:
            sys.stderr.write('\n'.join(err_msg))
            continue

        # ansible-doc returns plugins shipped with ansible-base using no namespace and collection.
        # For now, we fix these entries to use the ansible.builtin collection here.  The reason we
        # do it here instead of as part of a general normalization step is that other plugins
        # (site-specific ones from ANSIBLE_LIBRARY, for instance) will also be returned with no
        # collection name.  We know that we don't have any of those in this code (because we set
        # ANSIBLE_LIBRARY and other plugin path variables to /dev/null) so we can safely fix this
        # here but not outside the ansible-doc backend.
        fqcn = plugin_name
        try:
            get_fqcn_parts(fqcn)
        except ValueError:
            fqcn = f'ansible.builtin.{plugin_name}'

        results[fqcn] = json.loads(_filter_non_json_lines(stdout)[0])[plugin_name]

    return results


async def get_ansible_plugin_info(venv: Union['VenvRunner', 'FakeVenvRunner'],
                                  collection_dir: str) -> Dict[str, Dict[str, Any]]:
    """
    Retrieve information about all of the Ansible Plugins.

    :arg venv: A VenvRunner into which Ansible has been installed.
    :arg collection_dir: Directory in which the collections have been installed.
    :returns: A nested directory structure that looks like::

        plugin_type:
            plugin_name:  # Includes namespace and collection.
                {information from ansible-doc --json.  See the ansible-doc documentation for more
                 info.}
    """
    # Setup an sh.Command to run ansible-doc from the venv with only the collections we
    # found as providers of extra plugins.
    env = os.environ.copy()
    env.update(ANSIBLE_PATH_ENVVARS)
    env['ANSIBLE_COLLECTIONS_PATHS'] = collection_dir

    venv_ansible_doc = venv.get_command('ansible-doc')
    venv_ansible_doc = venv_ansible_doc.bake('-vvv', _env=env)

    extractors = {}
    for plugin_type in DOCUMENTABLE_PLUGINS:
        extractors[plugin_type] = asyncio.create_task(
            _get_plugin_info(plugin_type, venv_ansible_doc))

    results = await asyncio.gather(*extractors.values(), return_exceptions=True)

    plugin_map = {}
    err_msg = []
    an_exception = None
    for plugin_type, extraction_result in zip(extractors, results):

        if isinstance(extraction_result, Exception):
            an_exception = extraction_result
            formatted_exception = traceback.format_exception(None, extraction_result,
                                                             extraction_result.__traceback__)
            err_msg.append(f'Exception while parsing documentation for {plugin_type} plugins')
            err_msg.append(f'Exception:\n{"".join(formatted_exception)}')

        # Note: Exception will also be True.
        if isinstance(extraction_result, sh.ErrorReturnCode):
            stdout = extraction_result.stdout.decode("utf-8", errors="surrogateescape")
            stderr = extraction_result.stderr.decode("utf-8", errors="surrogateescape")
            err_msg.append(f'Full process stdout:\n{stdout}')
            err_msg.append(f'Full process stderr:\n{stderr}')

        if err_msg:
            sys.stderr.write('\n'.join(err_msg))
            sys.stderr.write('\n')
            continue

        plugin_map[plugin_type] = extraction_result

    if an_exception:
        # We wanted to print out all of the exceptions raised by parsing the output but once we've
        # done so, we want to then fail by raising one of the exceptions.
        raise ParsingError('Parsing of plugins failed')

    return plugin_map



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020

"""
Persist collection infornation used to build Ansible.

Build dependency files list the dependencies of an ACD release along with the
versions that are compatible with that release.

When we initially build an ACD major release, we'll use certain versions of collections.
We don't want to install backwards incompatible collections until the next major ACD release.
"""

from typing import TYPE_CHECKING, Dict, List, Mapping, NamedTuple, Optional

if TYPE_CHECKING:
    from packaging.version import Version as PyPiVersion
    from semantic_version import Version as SemVersion


class DependencyFileData(NamedTuple):
    ansible_version: str
    ansible_base_version: str
    deps: Dict[str, str]


class InvalidFileFormat(Exception):
    pass


def parse_pieces_file(pieces_file: str) -> List[str]:
    with open(pieces_file, 'rb') as f:
        contents = f.read()

    contents = contents.decode('utf-8')
    # One collection per line, ignoring comments and empty lines
    # TODO: PY3.8:
    # collections = [c for line in contents.split('\n')
    #                if (c := line.strip()) and not c.startswith('#')]
    collections = [line.strip() for line in contents.split('\n')
                   if line.strip() and not line.strip().startswith('#')]
    return collections


def _parse_name_version_spec_file(filename: str) -> DependencyFileData:
    deps: Dict[str, str] = {}
    ansible_base_version: Optional[str] = None
    acd_version: Optional[str] = None

    with open(filename, 'r') as f:
        for line in f:
            record = [entry.strip() for entry in line.split(':', 1)]

            if record[0] == '_acd_version':
                if acd_version is not None:
                    raise InvalidFileFormat(f'{filename} specified _acd_version'
                                            ' more than once')
                acd_version = record[1]
                continue

            if record[0] == '_ansible_base_version':
                if ansible_base_version is not None:
                    raise InvalidFileFormat(f'{filename} specified _ansible_base_version'
                                            ' more' ' than once')
                ansible_base_version = record[1]
                continue

            deps[record[0]] = record[1]

    if ansible_base_version is None or acd_version is None:
        raise InvalidFileFormat(f'{filename} was invalid.  It did not contain'
                                ' required fields')

    return DependencyFileData(acd_version, ansible_base_version, deps)


class DepsFile:
    """
    File containing the collections which are part of an Ansible release.

    A DepsFile contains a sequence of lines with a collection name, ": ", and then an exact
    version of the collection.  It tracks the exact collection-versions which were installed
    with a particular ansible version.

    The deps file has two special lines which are not collections.  They are::

        _ansible_version: X1.Y1.Z1
        _ansible_base_version: X2.Y2.Z2

    These are, respectively, the ansible version that was built and the ansible-base version which
    it was built against.  Note that the ansible release will depend on a compatible version of that
    ansible base version, not an exact dependency on that precise version.
    """

    def __init__(self, deps_file: str) -> None:
        """
        Create a :mod:`DepsFile`.

        :arg deps_file: filename of the `DepsFile`.
        """
        self.filename: str = deps_file

    def parse(self) -> DependencyFileData:
        """Parse the deps from a dependency file."""
        return _parse_name_version_spec_file(self.filename)

    def write(self, acd_version: str, ansible_base_version: str,
              included_versions: Mapping[str, str]) -> None:
        """
        Write a list of all the dependent collections included in this ACD release.

        :arg acd_version: The version of Ansible that is being recorded.
        :arg ansible_base_version: The version of Ansible base that will be depended on.
        :arg included_versions: Dictionary mapping collection names to the version range in this
            version of Ansible.
        """
        records = []
        for dep, version in included_versions.items():
            records.append(f'{dep}: {version}')
        records.sort()

        with open(self.filename, 'w') as f:
            f.write(f'_acd_version: {acd_version}\n')
            f.write(f'_ansible_base_version: {ansible_base_version}\n')
            f.write('\n'.join(records))
            f.write('\n')


class BuildFile:
    def __init__(self, build_file: str) -> None:
        self.filename: str = build_file

    def parse(self) -> DependencyFileData:
        """Parse the build from a dependency file."""
        return _parse_name_version_spec_file(self.filename)

    def write(self, acd_version: 'PyPiVersion', ansible_base_version: str,
              dependencies: Mapping[str, 'SemVersion']) -> None:
        """
        Write a build dependency file.

        A build dependency file records the collections that went into a given Ansible release along
        with the exact version of the collection that was included.

        :arg acd_version: The version of Ansible that is being recorded.
        :arg ansible_base_version: The version of Ansible base that will be depended on.
        :arg dependencies: Dictionary with keys of collection names and values of versions.
        """
        records = []
        for dep, version in dependencies.items():
            records.append(f'{dep}: >={version.major}.0.0,<{version.next_major()}')
        records.sort()

        with open(self.filename, 'w') as f:
            f.write(f'_acd_version: {acd_version.major}.{acd_version.minor}\n')
            f.write(f'_ansible_base_version: {ansible_base_version}\n')
            f.write('\n'.join(records))
            f.write('\n')



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Constant values for use throughout the antsibull codebase."""

from typing import FrozenSet


#: Number of bytes to read or write in one chunk
CHUNKSIZE = 4096

#: All the types of ansible plugins
PLUGIN_TYPES: FrozenSet[str] = frozenset(('become', 'cache', 'callback', 'cliconf', 'connection',
                                          'httpapi', 'inventory', 'lookup', 'shell', 'strategy',
                                          'vars', 'module', 'module_utils'))

#: The subset of PLUGINS which we build documentation for
DOCUMENTABLE_PLUGINS: FrozenSet[str] = frozenset(('become', 'cache', 'callback', 'cliconf',
                                                  'connection', 'httpapi', 'inventory', 'lookup',
                                                  'netconf', 'shell', 'vars', 'module',
                                                  'strategy',))

#: A good maximum number of concurrent threads to allow
THREAD_MAX: int = 128

#: A good maximum number of concurrent processes to allow
PROCESS_MAX: int = 7



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""
Functions for parsing fully qualified collection names.

Some definitions:

:Fully qualified collection names: aka FQCN are the complete name for a collection.  They
    include the namespace, collection name, and the plugin name.  If a plugin is within a python
    package within the collection dir, then the FQCN may include that as well.
:short_name: The pieces of the fqcn that don't include the namespace and collection name.  If the
    `collections:` keyword is used in an ansible playbook to get shorter names, the short name is
    what users would have to specify.
"""

import re
import typing as t


#: Format that a collection namespace and collection name must follow
NAMESPACE_RE_STR = '[a-z0-9][a-z0-9_]+'
#: Format of a FQCN
FQCN_RE = re.compile(fr'({NAMESPACE_RE_STR})\.({NAMESPACE_RE_STR})\.(.*)')


def get_fqcn_parts(fqcn: str) -> t.Tuple[str, str, str]:
    """
    Parse a fqcn into its three parts.

    :arg fqcn: The fqcn string.
    :returns: A tuple of (namespace, collection name, plugin short_name).
    :raises ValueError: If the string could not be parsed into fqcn components.
    """
    match = FQCN_RE.match(fqcn)
    if not match:
        raise ValueError(f'{fqcn} could not be parsed into fqcn parts')
    return match.groups()



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020

import asyncio
import os

import aiohttp
import semantic_version as semver

from .ansible_base import AnsibleBasePyPiClient
from .dependency_files import BuildFile, parse_pieces_file
from .galaxy import GalaxyClient


def display_exception(loop, context):
    print(context.get('exception'))


async def get_version_info(collections):
    loop = asyncio.get_running_loop()
    loop.set_exception_handler(display_exception)

    requestors = {}
    async with aiohttp.ClientSession() as aio_session:
        pypi_client = AnsibleBasePyPiClient(aio_session)
        requestors['_ansible_base'] = asyncio.create_task(pypi_client.get_versions())
        galaxy_client = GalaxyClient(aio_session)

        for collection in collections:
            requestors[collection] = asyncio.create_task(
                galaxy_client.get_versions(collection))

        collection_versions = {}
        responses = await asyncio.gather(*requestors.values())

        for idx, collection_name in enumerate(requestors):
            collection_versions[collection_name] = responses[idx]

    return collection_versions


def version_is_compatible(ansible_base_version, collection, version):
    # Metadata for this is not currently implemented.  So everything is rated as compatible
    return True


def find_latest_compatible(ansible_base_version, raw_dependency_versions):
    # Note: ansible-base compatibility is not currently implemented.  It will be a piece of
    # collection metadata that is present in the collection but may not be present in galaxy.  We'll
    # have to figure that out once the pieces are finalized

    # Order versions
    reduced_versions = {}
    for dep, versions in raw_dependency_versions.items():
        # Order the versions
        versions = [semver.Version(v) for v in versions]
        versions.sort(reverse=True)

        # Step through the versions to select the latest one which is compatible
        for version in versions:
            if version_is_compatible(ansible_base_version, dep, version):
                reduced_versions[dep] = version
                break

    return reduced_versions


def new_acd_command(args):
    collections = parse_pieces_file(args.pieces_file)
    dependencies = asyncio.run(get_version_info(collections))

    ansible_base_version = dependencies.pop('_ansible_base')[0]
    dependencies = find_latest_compatible(ansible_base_version, dependencies)

    build_filename = os.path.join(args.dest_dir, args.build_file)
    build_file = BuildFile(build_filename)
    build_file.write(args.acd_version, ansible_base_version, dependencies)

    return 0



import pytest

from antsibull import dependency_files as df

PIECES = """
community.general
# This is a comment
community.aws
    # Supported by ansible
    ansible.posix
    ansible.windows
# Third parties
purestorage.flasharray
"""

PARSED_PIECES = ['community.general', 'community.aws', 'ansible.posix', 'ansible.windows', 'purestorage.flasharray']

def test_parse_pieces(tmp_path):
    pieces_filename = tmp_path / 'pieces.in'
    with open(pieces_filename, 'w') as f:
        f.write(PIECES)
    assert df.parse_pieces_file(pieces_filename) == PARSED_PIECES



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Entrypoint to the antsibull-docs script."""

import argparse
import os.path
import stat
import sys
from typing import Callable, Dict, List

# import twiggy

# from ..config import load_config
from ..constants import DOCUMENTABLE_PLUGINS
from ..filesystem import writable_via_acls
from .doc_commands import collection, devel, plugin, stable, current

#: Mapping from command line subcommand names to functions which implement those
#: The functions need to take a single argument, the processed list of args.
ARGS_MAP: Dict[str, Callable] = {'devel': devel.generate_docs,
                                 'stable': stable.generate_docs,
                                 'current': current.generate_docs,
                                 'collection': collection.generate_docs,
                                 'plugin': plugin.generate_docs,
                                 }

#: The filename for the file which lists raw collection names
DEFAULT_PIECES_FILE: str = 'acd.in'


class InvalidArgumentError(Exception):
    """A problem parsing or validating a command line argument."""


def _normalize_common_options(args: argparse.Namespace) -> None:
    if args.command is None:
        raise InvalidArgumentError('Please specify a subcommand to run')

    args.dest_dir = os.path.expanduser(os.path.expandvars(args.dest_dir))
    args.dest_dir = os.path.abspath(os.path.realpath(args.dest_dir))

    # We're going to be writing a deep hierarchy of files into this directory so we need to make
    # sure that the user understands that this needs to be a directory which has been secured
    # against malicious usage:
    stat_results = os.stat(args.dest_dir)

    # Exists already
    if not stat.S_ISDIR(stat_results.st_mode):
        raise InvalidArgumentError(f'{args.dest_dir} must be an existing directory')

    # Owned by the user
    euid = os.geteuid()
    if stat_results[stat.ST_UID] != euid:
        raise InvalidArgumentError(f'{args.dest_dir} must be owned by you')

    # Writable only by the user
    if stat.S_IMODE(stat_results.st_mode) & (stat.S_IWOTH | stat.S_IWGRP):
        raise InvalidArgumentError(f'{args.dest_dir} must only be writable by the owner')

    if writable_via_acls(args.dest_dir, euid):
        raise InvalidArgumentError(f'Filesystem acls grant write on {args.dest_dir} to'
                                   ' additional users')


def _normalize_devel_options(args: argparse.Namespace) -> None:
    if args.command != 'devel':
        return

    if not os.path.isfile(args.pieces_file):
        raise InvalidArgumentError(f'The pieces file, {args.pieces_file}, must already exist.'
                                   ' It should contain one namespace.collection per line')


def _normalize_stable_options(args: argparse.Namespace) -> None:
    if args.command != 'stable':
        return

    if not os.path.isfile(args.deps_file):
        raise InvalidArgumentError(f'The deps file, {args.deps_file}, must already exist.'
                                   ' It should contain one namespace.collection with version'
                                   ' per line')


def _normalize_current_options(args: argparse.Namespace) -> None:
    if args.command != 'current':
        return

    if not os.path.isdir(args.collection_dir) or not os.path.isdir(
            os.path.join(args.collection_dir, 'ansible_collections')):
        raise InvalidArgumentError(f'The collection directory, {args.collection_dir}, must be'
                                   ' a directory containing a subdirectory ansible_collections')


def _normalize_plugin_options(args: argparse.Namespace) -> None:
    if args.command != 'plugin':
        return

    if not os.path.isfile(args.plugin):
        raise InvalidArgumentError(f'The plugin file, {args.plugin}, must exist.')


def parse_args(program_name: str, args: List[str]) -> argparse.Namespace:
    """
    Parse and coerce the command line arguments.

    :arg program_name: The name of the program
    :arg args: A list of the command line arguments
    :returns: A :python:obj:`argparse.Namespace`
    :raises InvalidArgumentError: Whenever there's something wrong with the arguments.
    """
    # TODO: Need a function to return a parser with options that all antsibull
    # scripts use. Then we can add it as a parent to the common_parser.  First use case:
    # config file.
    # antsibull_parser =

    common_parser = argparse.ArgumentParser(add_help=False)
    common_parser.add_argument('--dest-dir', default='.',
                               help='Directory to write the output to')

    cache_parser = argparse.ArgumentParser(add_help=False)
    cache_parser.add_argument('--ansible-base-cache', default=None,
                              help='Checkout or expanded tarball of the ansible-base package.  If'
                              ' this is a git checkout it must be the HEAD of the cache branch.'
                              ' If it is an expanded tarball, the __version__ will be checked to'
                              ' make sure it is compatible with and the same or later version than'
                              ' requested by the depcs file.')
    cache_parser.add_argument('--collection-cache', default=None,
                              help='Directory of collection tarballs.  These will be used instead'
                              ' of downloading fresh versions provided that they meet the criteria'
                              ' (Latest version of the collections known to galaxy).')

    parser = argparse.ArgumentParser(prog=program_name,
                                     description='Script to manage generated documentation for'
                                     ' ansible')
    subparsers = parser.add_subparsers(title='Subcommands', dest='command',
                                       help='for help use  SUBCOMMANDS -h')

    # Document the next version of ansible
    devel_parser = subparsers.add_parser('devel', parents=[common_parser, cache_parser],
                                         description='Generate documentation for the next major'
                                         ' release of Ansible')
    devel_parser.add_argument('--pieces-file', default=DEFAULT_PIECES_FILE,
                              help='File containing a list of collections to include')

    stable_parser = subparsers.add_parser('stable',
                                          parents=[common_parser, cache_parser],
                                          description='Generate documentation for a current'
                                          ' version of ansible')
    stable_parser.add_argument('--deps-file', required=True,
                               help='File which contains the list of collections and'
                               ' versions which were included in this version of Ansible')

    current_parser = subparsers.add_parser('current',
                                           parents=[common_parser],
                                           description='Generate documentation for the current'
                                           ' installed version of ansible and the current installed'
                                           ' collections')
    current_parser.add_argument('--collection-dir', required=True,
                                help='Path to the directory containing ansible_collections')

    collection_parser = subparsers.add_parser('collection',
                                              parents=[common_parser],
                                              description='Generate documentation for a single'
                                              ' collection')
    collection_parser.add_argument('--collection-version', default='@latest',
                                   help='The version of the collection to document.  The special'
                                   ' version, "@latest" can be used to download and document the'
                                   ' latest version from galaxy')
    collection_parser.add_argument(nargs='+', dest='collections', action='append', default=[],
                                   help='One or more collections to document.  If the names are'
                                   ' directories on disk, they will be parsed as expanded'
                                   ' collections. Otherwise, if they could be collection'
                                   ' names, they will be downloaded from galaxy')

    file_parser = subparsers.add_parser('plugin',
                                        parents=[common_parser],
                                        description='Generate documentation for a single plugin')
    file_parser.add_argument(nargs=1, dest='plugin', action='store',
                             choices=DOCUMENTABLE_PLUGINS,
                             help='A single file to document.')
    file_parser.add_argument('--plugin-type', action='store', default='module',
                             help='The type of the plugin')

    args: argparse.Namespace = parser.parse_args(args)

    # Validation and coercion
    _normalize_common_options(args)
    _normalize_devel_options(args)
    _normalize_stable_options(args)
    _normalize_current_options(args)
    _normalize_plugin_options(args)

    # Note: collections aren't validated as existing files or collection names here because talking
    # to galaxy to validate the collection names goes beyond the scope of what parsing and
    # validating the command line should do.

    return args


def run(args: List[str]) -> int:
    """
    Run the program.

    :arg args: A list of command line arguments.  Typically :python:`sys.argv`.
    :returns: A program return code.  0 for success, integers for any errors.  These are documented
        in :func:`main`.
    """
    program_name = os.path.basename(args[0])
    try:
        args: argparse.Namespace = parse_args(program_name, args[1:])
    except InvalidArgumentError as e:
        print(e)
        return 2

    # Need to finish implementing config loading
    # cfg = load_config(args.cfg_file)
    # if cfg['logging_cfg']:
    #    twiggy.dict_config(cfg['logging_cfg'])

    return ARGS_MAP[args.command](args)


def main() -> int:
    """
    Entrypoint called from the script.

    console_scripts call functions which take no parameters.  However, it's hard to test a function
    which takes no parameters so this function lightly wraps :func:`run`, which actually does the
    heavy lifting.

    :returns: A program return code.

    Return codes:
        :0: Success
        :1: Unhandled error.  See the Traceback for more information.
        :2: There was a problem with the command line arguments
        :3: Unexpected problem downloading ansible-base
    """
    return run(sys.argv)



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Functionality to work with a venv."""

import os
import venv

import sh


class VenvRunner:
    """
    Makes running a command in a venv easy.

    Combines venv functionality with sh.

    .. seealso::
        * `sh <https://amoffat.github.io/sh/>`_
        * :python:mod:`venv`
    """

    def __init__(self, name: str, top_dir: str) -> None:
        """
        Create a venv.

        :arg name: Name of the venv.
        :arg top_dir: Directory the venv will be created inside of.
        """
        self.name = name
        self.top_dir = top_dir
        self.venv_dir: str = os.path.join(top_dir, name)
        venv.create(self.venv_dir, clear=True, symlinks=True, with_pip=True)
        self._python = self.get_command('python')

    def get_command(self, executable_name) -> sh.Command:
        """
        Return an :sh:obj:`sh.Command` for the given program installed within the venv.

        :arg executable_name: Program to return a command for.
        :returns: An :obj:`sh.Command` that will invoke the program.
        """
        return sh.Command(os.path.join(self.venv_dir, 'bin', executable_name))

    def install_package(self, package_name: str, from_project_path=False) -> sh.RunningCommand:
        """
        Install a python package into the venv.

        :arg package_name: This can be a bare package name or a path to a file.  It's passed
            directly to :command:`pip install`.
        :arg from_project_path: Instead of a package name or file, package_name is a path to a
            project.  ie: an expanded sdist or a checkout of a source repo.  At the moment,
            pip -e is used to install these sorts of packages but this is an implementation
            detail.
        :returns: An :sh:obj:`sh.RunningCommand` for the pip output.
        """
        if from_project_path:
            package_args = ('-e', package_name)
        else:
            package_args = (package_name,)

        return self._python('-m', 'pip', 'install', *package_args)


class FakeVenvRunner:
    """
    Simply runs commands.

    .. seealso::
        * `sh <https://amoffat.github.io/sh/>`_
        * :python:mod:`venv`
    """

    def get_command(self, executable_name) -> sh.Command:
        """
        Return an :sh:obj:`sh.Command` for the given program installed within the venv.

        :arg executable_name: Program to return a command for.
        :returns: An :obj:`sh.Command` that will invoke the program.
        """
        return sh.Command(executable_name)

    def install_package(self, package_name: str, from_project_path=False) -> sh.RunningCommand:
        """
        Install a python package into the venv.

        :arg package_name: This can be a bare package name or a path to a file.  It's passed
            directly to :command:`pip install`.
        :arg from_project_path: Instead of a package name or file, package_name is a path to a
            project.  ie: an expanded sdist or a checkout of a source repo.  At the moment,
            pip -e is used to install these sorts of packages but this is an implementation
            detail.
        :returns: An :sh:obj:`sh.RunningCommand` for the pip output.
        """
        raise Exception('Not implemented')



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020

import asyncio
from datetime import datetime
import os
import os.path
import pkgutil
import shutil
import tempfile
from functools import partial

import aiofiles
import aiohttp
import sh
from jinja2 import Template

from .collections import install_separately, install_together
from .dependency_files import BuildFile, DepsFile
from .galaxy import CollectionDownloader


#
# Common code
#

async def download_collections(deps, download_dir):
    requestors = {}
    async with aiohttp.ClientSession() as aio_session:
        downloader = CollectionDownloader(aio_session, download_dir)
        for collection_name, version_spec in deps.items():
            requestors[collection_name] = asyncio.create_task(
                downloader.download_latest_matching(collection_name, version_spec))

        included_versions = {}
        responses = await asyncio.gather(*requestors.values())

    # Note: Python dicts have a stable sort order and since we haven't modified the dict since we
    # used requestors.values() to generate responses, requestors and responses therefor have
    # a matching order.
    for collection_name, results in zip(requestors, responses):
        included_versions[collection_name] = results.version

    return included_versions


#
# Single sdist for ansible
#

def copy_boilerplate_files(package_dir):
    gpl_license = pkgutil.get_data('antsibull.data', 'gplv3.txt')
    with open(os.path.join(package_dir, 'COPYING'), 'wb') as f:
        f.write(gpl_license)

    readme = pkgutil.get_data('antsibull.data', 'acd-readme.txt')
    with open(os.path.join(package_dir, 'README'), 'wb') as f:
        f.write(readme)


def write_manifest(package_dir, debian=False):
    manifest_file = os.path.join(package_dir, 'MANIFEST.in')
    with open(manifest_file, 'w') as f:
        f.write('include COPYING\n')
        f.write('include README\n')
        if debian:
            f.write('include debian/*\n')
        f.write('recursive-include ansible_collections/ **\n')


def write_setup(acd_version, collection_deps, package_dir):
    setup_filename = os.path.join(package_dir, 'setup.py')

    setup_tmpl = Template(pkgutil.get_data('antsibull.data', 'acd-setup_py.j2').decode('utf-8'))
    setup_contents = setup_tmpl.render(version=acd_version, collection_deps=collection_deps)

    with open(setup_filename, 'w') as f:
        f.write(setup_contents)


def write_python_build_files(acd_version, collection_deps, package_dir, debian=False):
    copy_boilerplate_files(package_dir)
    write_manifest(package_dir, debian)
    write_setup(acd_version, collection_deps, package_dir)


def write_debian_directory(acd_version, package_dir):
    debian_dir = os.path.join(package_dir, 'debian')
    os.mkdir(debian_dir, mode=0o700)
    debian_files = ('changelog.j2', 'control', 'copyright', 'rules')
    for filename in debian_files:
        # Don't use os.path.join here, the get_data docs say it should be
        # slash-separated.
        src_pkgfile = 'debian/' + filename
        data = pkgutil.get_data('antsibull.data', src_pkgfile).decode('utf-8')

        if filename.endswith('.j2'):
            filename = filename.replace('.j2', '')
            # If the file is a template, send it in vars it might need
            # and update 'data' to be the result.
            tmpl = Template(data)
            data = tmpl.render(
                version=acd_version,
                date=datetime.utcnow().strftime("%a, %d %b %Y %T +0000"),
            )

        with open(os.path.join(debian_dir, filename), 'w') as f:
            f.write(data)


def make_dist(ansible_dir, dest_dir):
    sh.python('setup.py', 'sdist', _cwd=ansible_dir)
    dist_dir = os.path.join(ansible_dir, 'dist')
    files = os.listdir(dist_dir)
    if len(files) != 1:
        raise Exception('python setup.py sdist should only have created one file')

    shutil.move(os.path.join(dist_dir, files[0]), dest_dir)


def build_single_command(args):
    build_file = BuildFile(args.build_file)
    build_acd_version, ansible_base_version, deps = build_file.parse()

    if not str(args.acd_version).startswith(build_acd_version):
        print(f'{args.build_file} is for version {build_acd_version} but we need'
              ' {args.acd_version.major}.{arg.acd_version.minor}')
        return 1

    with tempfile.TemporaryDirectory() as tmp_dir:
        download_dir = os.path.join(tmp_dir, 'collections')
        os.mkdir(download_dir, mode=0o700)

        included_versions = asyncio.run(download_collections(deps, download_dir))

        package_dir = os.path.join(tmp_dir, f'ansible-{args.acd_version}')
        os.mkdir(package_dir, mode=0o700)
        ansible_collections_dir = os.path.join(package_dir, 'ansible_collections')
        os.mkdir(ansible_collections_dir, mode=0o700)

        collections_to_install = [p for f in os.listdir(download_dir)
                                  if os.path.isfile(p := os.path.join(download_dir, f))]
        asyncio.run(install_together(collections_to_install, ansible_collections_dir))
        write_python_build_files(args.acd_version, '', package_dir, args.debian)
        if args.debian:
            write_debian_directory(args.acd_version, package_dir)
        make_dist(package_dir, args.dest_dir)

    deps_filename = os.path.join(args.dest_dir, args.deps_file)
    deps_file = DepsFile(deps_filename)
    deps_file.write(args.acd_version, ansible_base_version, included_versions)

    return 0

#
# Code to make one sdist per collection
#


async def write_collection_readme(collection_name, package_dir):
    readme_tmpl = Template(pkgutil.get_data('antsibull.data',
                                            'collection-readme.j2').decode('utf-8'))
    readme_contents = readme_tmpl.render(collection_name=collection_name)

    readme_filename = os.path.join(package_dir, 'README.rst')
    async with aiofiles.open(readme_filename, 'w') as f:
        await f.write(readme_contents)


async def write_collection_setup(name, version, package_dir):
    setup_filename = os.path.join(package_dir, 'setup.py')

    setup_tmpl = Template(pkgutil.get_data('antsibull.data',
                                           'collection-setup_py.j2').decode('utf-8'))
    setup_contents = setup_tmpl.render(version=version, name=name)

    async with aiofiles.open(setup_filename, 'w') as f:
        await f.write(setup_contents)


async def write_collection_manifest(package_dir):
    manifest_file = os.path.join(package_dir, 'MANIFEST.in')
    async with aiofiles.open(manifest_file, 'w') as f:
        await f.write('include README.rst\n')
        await f.write('recursive-include ansible_collections/ **\n')


async def make_collection_dist(name, version, package_dir, dest_dir):
    # Copy boilerplate into place
    await write_collection_readme(name, package_dir)
    await write_collection_setup(name, version, package_dir)
    await write_collection_manifest(package_dir)

    loop = asyncio.get_running_loop()

    # Create the python sdist
    await loop.run_in_executor(None, partial(sh.python, 'setup.py', 'sdist', _cwd=package_dir))
    dist_dir = os.path.join(package_dir, 'dist')
    files = os.listdir(dist_dir)
    if len(files) != 1:
        raise Exception('python setup.py sdist should only have created one file')

    dist_file = os.path.join(dist_dir, files[0])
    await loop.run_in_executor(None, shutil.move, dist_file, dest_dir)


async def make_collection_dists(dest_dir, collection_dirs):
    dist_creators = []
    for collection_dir in collection_dirs:
        dir_name_only = os.path.basename(collection_dir)
        dummy_, dummy_, name, version = dir_name_only.split('-', 3)

        dist_creators.append(asyncio.create_task(
            make_collection_dist(name, version, collection_dir, dest_dir)))

    await asyncio.gather(*dist_creators)


def build_multiple_command(args):
    build_file = BuildFile(args.build_file)
    build_acd_version, ansible_base_version, deps = build_file.parse()

    if not str(args.acd_version).startswith(build_acd_version):
        print(f'{args.build_file} is for version {build_acd_version} but we need'
              f' {args.acd_version.major}.{args.acd_version.minor}')
        return 1

    with tempfile.TemporaryDirectory() as tmp_dir:
        download_dir = os.path.join(tmp_dir, 'collections')
        os.mkdir(download_dir, mode=0o700)

        included_versions = asyncio.run(download_collections(deps, download_dir))
        collections_to_install = [p for f in os.listdir(download_dir)
                                  if os.path.isfile(p := os.path.join(download_dir, f))]
        collection_dirs = asyncio.run(install_separately(collections_to_install, download_dir))
        asyncio.run(make_collection_dists(args.dest_dir, collection_dirs))

        # Create the ansible package that deps on the collections we just wrote
        package_dir = os.path.join(tmp_dir, f'ansible-{args.acd_version}')
        os.mkdir(package_dir, mode=0o700)
        ansible_collections_dir = os.path.join(package_dir, 'ansible_collections')
        os.mkdir(ansible_collections_dir, mode=0o700)

        # Construct the list of dependent collection packages
        collection_deps = []
        for collection, version in sorted(included_versions.items()):
            collection_deps.append(f"        '{collection}>={version},<{version.next_major()}'")
        collection_deps = '\n' + ',\n'.join(collection_deps)
        write_python_build_files(args.acd_version, collection_deps, package_dir)

        make_dist(package_dir, args.dest_dir)

    # Write the deps file
    deps_filename = os.path.join(args.dest_dir, args.deps_file)
    deps_file = DepsFile(deps_filename)
    deps_file.write(args.acd_version, ansible_base_version, included_versions)

    return 0



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020

import json
import os.path
import pkgutil
import tempfile

import sh
from jinja2 import Template

from .dependency_files import DepsFile


def build_collection_command(args):
    with tempfile.TemporaryDirectory() as working_dir:
        collection_dir = os.path.join(working_dir, 'community', 'acd')

        sh.ansible_galaxy('collection', 'init', 'community.acd', '--init-path', working_dir)
        # Copy the README.md file
        readme = pkgutil.get_data('antsibull.data', 'README_md.txt')
        with open(os.path.join(collection_dir, 'README.md'), 'wb') as f:
            f.write(readme)

        # Parse the deps file
        deps_file = DepsFile(args.deps_file)
        dummy1_, dummy2_, deps = deps_file.parse()

        # Template the galaxy.yml file
        dep_string = json.dumps(deps)
        dep_string.replace(', ', ',\n    ')
        galaxy_yml = pkgutil.get_data('antsibull.data', 'galaxy_yml.j2').decode('utf-8')
        galaxy_yml_tmpl = Template(galaxy_yml)
        galaxy_yml_contents = galaxy_yml_tmpl.render(version=args.acd_version,
                                                     dependencies=dep_string)

        with open(os.path.join(collection_dir, 'galaxy.yml'), 'w') as f:
            f.write(galaxy_yml_contents)

        sh.ansible_galaxy('collection', 'build', '--output-path', args.dest_dir, collection_dir)

    return 0



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Functions to setup logging."""
import twiggy

# Temporarily setup twiggy logging with defaults until we can get real configuration.
twiggy.quick_setup()

#: The standard log to use everywhere.  The name of the logger reflects the project that
#: the logger is used within.  For those used to using the module's __name__ field as the name, the
#: idiom we use here is to set the module name in the `mod` field.
log = twiggy.log.name('antsibull')
log.min_level = twiggy.levels.DISABLED

# This is how to setup the logger to use in your module
mlog = log.fields(mod=__name__)
mlog.debug('logging loaded')

__all__ = ('log',)



"""
Test the set of Schemas altogether to see that they will parse information correctly.
"""
import glob
import json
import os.path

import pytest

from antsibull.schemas import ansible_doc as ad


SINGLE_TESTS = {
    'one_become.json': ad.BecomePluginSchema,
    'one_cache.json': ad.CachePluginSchema,
    'one_callback.json': ad.CallbackPluginSchema,
    'one_cliconf.json': ad.CliConfPluginSchema,
    'one_connection.json': ad.ConnectionPluginSchema,
    'one_httpapi.json': ad.HttpApiPluginSchema,
    'one_inventory.json': ad.InventoryPluginSchema,
    'one_lookup.json': ad.LookupPluginSchema,
    'one_module.json': ad.ModulePluginSchema,
    'one_netconf.json': ad.NetConfPluginSchema,
    'one_shell.json': ad.ShellPluginSchema,
    'one_strategy.json': ad.StrategyPluginSchema,
    'one_vars.json': ad.VarsPluginSchema,
}


@pytest.mark.parametrize('test_file, test_schema', SINGLE_TESTS.items())
def test_one_plugin_of_each_type(test_file, test_schema):
    plugin_type = os.path.splitext(os.path.basename(test_file))[0]
    plugin_type = plugin_type[len('one_'):]

    test_dir = os.path.dirname(__file__)
    result_file = os.path.join(test_dir, 'good_data', 'one_%s_results.json' % plugin_type)
    full_path = os.path.join(test_dir, 'good_data', test_file)

    with open(result_file, 'r') as f:
        results = json.load(f)

    with open(full_path, 'r') as f:
        ansible_doc_output = f.read()

    model = test_schema.parse_raw(ansible_doc_output)

    model_dict = model.dict()
    assert model_dict == results



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Compat for older versions of Python."""

import asyncio
import sys


def _dummy():
    pass


if sys.version_info < (3, 7):
    #: In Python3.7+, :python:meth:`asyncio.get_running_loop` is the preferred way to get the loop.
    #: It gets the loop currently running.  When that's not available, we can use
    #: :meth:`asyncio.get_event_loop` which gets the default loop.  These are close to the same in
    #: our code since we use the default loop.  Just remember not to run it outside of an event
    #: loop.
    best_get_loop = asyncio.get_event_loop
    _loop = asyncio.get_event_loop()
    #: In Python3.7+, :python:meth:`asyncio.run` is the preferred way to enter an asyncio event loop
    #: and start to execute asynchronous code.  In earlier Pythons,
    #: :python:meth:`run_until_complete` can be used in almost the same manner since we are using
    #: the default event loop.
    asyncio_run = _loop.run_until_complete
else:
    best_get_loop = asyncio.get_running_loop
    asyncio_run = asyncio.run


__all__ = ('asyncio_run', 'best_get_loop')



# Copyright: (c) 2019, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import re

try:
    from html import escape as html_escape
except ImportError:
    # Python-3.2 or later
    import cgi

    def html_escape(text, quote=True):
        return cgi.escape(text, quote)

from jinja2.runtime import Undefined


_ITALIC = re.compile(r"I\(([^)]+)\)")
_BOLD = re.compile(r"B\(([^)]+)\)")
_MODULE = re.compile(r"M\(([^)]+)\)")
_URL = re.compile(r"U\(([^)]+)\)")
_LINK = re.compile(r"L\(([^)]+), *([^)]+)\)")
_REF = re.compile(r"R\(([^)]+), *([^)]+)\)")
_CONST = re.compile(r"C\(([^)]+)\)")
_RULER = re.compile(r"HORIZONTALLINE")


def html_ify(text):
    ''' convert symbols like I(this is in italics) to valid HTML '''

    t = html_escape(text)
    t = _ITALIC.sub(r"<em>\1</em>", t)
    t = _BOLD.sub(r"<b>\1</b>", t)
    t = _MODULE.sub(r"<span class='module'>\1</span>", t)
    t = _URL.sub(r"<a href='\1'>\1</a>", t)
    t = _REF.sub(r"<span class='module'>\1</span>", t)
    t = _LINK.sub(r"<a href='\2'>\1</a>", t)
    t = _CONST.sub(r"<code>\1</code>", t)
    t = _RULER.sub(r"<hr/>", t)

    return t.strip()


def documented_type(text):
    ''' Convert any python type to a type for documentation '''

    if isinstance(text, Undefined):
        return '-'
    if text == 'str':
        return 'string'
    if text == 'bool':
        return 'boolean'
    if text == 'int':
        return 'integer'
    if text == 'dict':
        return 'dictionary'
    return text


# The max filter was added in Jinja2-2.10.  Until we can require that version, use this
def do_max(seq):
    return max(seq)


def rst_ify(text):
    ''' convert symbols like I(this is in italics) to valid restructured text '''

    t = _ITALIC.sub(r"*\1*", text)
    t = _BOLD.sub(r"**\1**", t)
    t = _MODULE.sub(r":ref:`\1 <\1_module>`", t)
    t = _LINK.sub(r"`\1 <\2>`_", t)
    t = _URL.sub(r"\1", t)
    t = _REF.sub(r":ref:`\1 <\2>`", t)
    t = _CONST.sub(r"``\1``", t)
    t = _RULER.sub(r"------------", t)

    return t


def rst_fmt(text, fmt):
    ''' helper for Jinja2 to do format strings '''

    return fmt % (text)


def rst_xline(width, char="="):
    ''' return a restructured text line of a given length '''

    return char * width



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['antsibull',
 'antsibull.cli',
 'antsibull.cli.doc_commands',
 'antsibull.data',
 'antsibull.data.docsite',
 'antsibull.docs_parsing',
 'antsibull.jinja2',
 'antsibull.schemas',
 'antsibull.vendored',
 'tests',
 'tests.functional.schema',
 'tests.units']

package_data = \
{'': ['*'],
 'antsibull.data': ['debian/*'],
 'tests.functional.schema': ['good_data/*']}

install_requires = \
['PyYAML',
 'aiofiles',
 'aiohttp',
 'antsibull-changelog',
 'asyncio-pool',
 'docutils',
 'jinja2',
 'packaging',
 'perky',
 'pydantic',
 'rstcheck>=3,<4',
 'semantic_version',
 'sh',
 'twiggy']

entry_points = \
{'console_scripts': ['antsibull-build = antsibull.cli.antsibull_build:main',
                     'antsibull-docs = antsibull.cli.antsibull_docs:main',
                     'antsibull-lint = antsibull.cli.antsibull_lint:main']}

setup_kwargs = {
    'name': 'antsibull',
    'version': '0.7.0',
    'description': 'Tools for building the Ansible Distribution',
    'long_description': "# antsibull -- Ansible Build Scripts\nTooling for building various things related to Ansible\n\nScripts that are here:\n\n* antsibull-build - Builds Ansible-2.10+ from component collections ([docs](docs/build-ansible.rst))\n* antsibull-docs - Extracts documentation from ansible plugins\n* antsibull-lint - Right now only validates ``changelogs/changelog.yaml`` files ([docs](docs/changelog.yaml-format.md))\n\nA related project is [antsibull-changelog](https://pypi.org/project/antsibull-changelog/), which is in its [own repository](https://github.com/ansible-community/antsibull-changelog/).\n\nScripts are created by poetry at build time.  So if you want to run from\na checkout, you'll have to run them under poetry::\n\n    python3 -m pip install poetry\n    poetry install  # Installs dependencies into a virtualenv\n    poetry run antsibull-build --help\n\nIf you want to create a new release::\n\n    poetry build\n    poetry publish  # Uploads to pypi.  Be sure you really want to do this\n\n.. note:: When installing a package published by poetry, it is best to use\n    pip >= 19.0.  Installing with pip-18.1 and below could create scripts which\n    use pkg_resources which can slow down startup time (in some environments by\n    quite a large amount).\n\nUnless otherwise noted in the code, it is licensed under the terms of the GNU\nGeneral Public License v3 or, at your option, later.\n",
    'author': 'Toshio Kuratomi',
    'author_email': 'a.badger@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ansible-community/antsibull',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.0,<4.0.0',
}


setup(**setup_kwargs)



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020

# Ignore Unitialized attribute errors because BaseModel works some magic
# to initialize the attributes when data is loaded into them.
# pyre-ignore-all-errors[13]
"""
Base schemas and supporting validator functions.

`Pydantic <https://pydantic-docs.helpmanual.io/>`_  is written to be an advanced version of
:python:mod:`dataclasses`.  It's primary use case is creating Python objects that hold data (think
of traditional C-style ``structs``).  As such, it is optimized for a slightly different use case
than what we are using it for.  However, the utility methods that come with pydantic make it quite
usable for data validation and normalization even with this difference in goals.

How to use these schemas effectively
====================================

The schemas shipped with antsibull allow you to validate and normalize plugin documentation.  This
way, you can use data from the plugins for web documentation, display on the command line, or
testing that your documentation is well formed.

To use these schemas, follow these steps:

* Choose the Schema that matches with the level of data that you want to validate and normalize.

    * If you want to validate and normalize :ansible:cmd:`ansible-doc` output, use the schemas in
      :mod:`antsibull.schemas.ansible_doc`.

      * :obj:`~antsibull.schemas.ansible_doc.AnsibleDocSchema` lets you validate the documentation
        for multiple plugins at once.  It is useful if all you want to do is validate documentation
        as you can run it once and then get all the errors.

      * If you want to normalize the data and use it to make documentation then the schemas in
        :attr:`~antsibull.schemas.ansible_doc.ANSIBLE_DOC_SCHEMAS` might be more appropriate.
        These schemas can be run on individual plugin data.  The advantage of using them is that
        an error in documentation will only fail that one plugin, not all of them.

    * If you need more fine grained control, you can use the schemas in
      :mod:`antsibull.schemas.docs.DOCS_SCHEMAS`.  These schemas give you access to the individual
      components of a plugin's documentation, doc, example, metadata, and return.  That way you can
      create documentation if any one of these (or specific ones) can be normalized even if the
      others cannot.

* Use one of the pydantic.BaseModel `alternate constructors
  <https://pydantic-docs.helpmanual.io/usage/models/#helper-functions>`_
  to load the data into the Schema.  :ansible:cmd:`ansible-doc` output is json, for instance, so you
  can feed that directly to :meth:`ansible_doc.ModulePluginSchema.parse_raw` (or the same method on
  a different schema).  If you have to manipulate the data as a dict first, you can use
  :meth:`ansible_doc.ModulePluginSchema.parse_obj`.

* The Schema will validate and normalize the data as it is loaded.

* Call the `dict() <https://pydantic-docs.helpmanual.io/usage/exporting_models/#modeldict>`_
  method on the returned model to convert the data back into a dict so that you can use it with
  templating engines or modify the structure of the data.

One example of doing all this:

.. code-block:: pycon

    >>> import sh, jinja2
    >>> from antsibull.schemas import ansible_doc
    >>> template = jinja2.Template('{{ name }} -- {{ doc["short_description"] }}')
    >>> module_json = sh.ansible_doc('-t', 'module', '--json', 'yum').stdout
    >>> module_model = ansible_doc.ModulePluginSchema.parse_raw(module_json)
    >>> # Retrieve the data from the __root__ field of the dict representation:
    >>> module_dict = module_model.dict()['__root__']
    >>> for plugin_name, plugin_info in module_dict.items():
    >>>     out = template.render(name=plugin_name, doc=plugin_info['doc'])
    >>>     print(out)
    >>> # yum -- Manages packages with the I(yum) package manager

Writing good pydantic schemas
=============================

* If you encounter something that needs to allow a certain value for legacy reasons but you don't
  want people to do that in the future, put the desired behaviour into the model and put the
  unwanted behaviour into a validator.  This makes it so the json-schema that can be built from
  the model will encapsulate the desired behaviour so that when people check their documentation
  with that, they will get an error while the code using the pydantic validator directly will
  automatically correct the data.

* The corollary is that when the definition should be added to the json-schema (for instance,
  to make validation more strict), those should be added as part of the type information in
  the model.

An example of both of the above:

.. code-block:: python

      class Model(BaseModel):
          new_name: str
          # Use pydantic.constr so that json-schema validations will also need to match the regex
          type: constr(regex='^(int|str|float)$')

          # Use validator on a specific attribute to change the value of the attribute
          @validator('type', pre=True)
          def fix_alternate_type_names(cls, value):
              '''These type names are commonly typoed.'''
              if value == 'string':
                  return 'str'
              if value == 'integer':
                  return 'int'
              return value

          # Use a root_validator to rename an attribute name to something else
          @root_validator(pre=True)
          def handle_renamed_attribute(cls, values):
              '''Rename an attribute to a new name.'''
              old_name = values.get('old_name')
              if old_name:
                  if values.get('new_name'):
                      raise ValueError('Only use `new_name`, not `old_name`.')
                  values['new_name'] = old_name
                  del values['old_name']
              return values

"""
import typing as t
from collections.abc import Mapping

import pydantic as p
import yaml

_SENTINEL = object()


#: Python scalars that are represented in JSON
JSONScalarT = t.Union[None, int, float, str]

#: Python types represented in JSON.  Unfortunately, we can't define this recursively yet.
JSONValueT = t.Union[JSONScalarT, t.List[t.Any], t.Dict[str, t.Any]]

#: Constrained string for valid Ansible environment variables
REQUIRED_ENV_VAR_F = p.Field(..., regex='[A-Z_]+')

#: option types are a set of strings that represent the types handled by argspec.
OPTION_TYPE_F = p.Field('str', regex='^(bits|bool|bytes|dict|float|int|json|jsonarg|list'
                        '|path|raw|sid|str)$')

#: Constrained string type for version numbers
REQUIRED_VERSION_F = p.Field(..., regex='^([0-9][0-9.]+)$')

#: Constrained string listing the possible types of a return field
RETURN_TYPE_F = p.Field('str', regex='^(bool|complex|dict|float|int|list|str)$')


def list_from_scalars(obj):
    """
    Create a list if the obj is a select scalary object.

    obj must be int, str, float, or None to be converted to a list.  Note that None converts to
    the empty list, rather than a list with None as its sole element.

    :arg obj: The object to convert if necessary.
    :return: obj wrapped in a list or the list itself.
    """
    if isinstance(obj, (str, int, float)):
        return [obj]

    if obj is None:
        return []

    return obj


def transform_return_docs(obj):
    """
    Attempt to convert data to a dict using a known strategy otherwise return.

    :arg obj: The return doc field to try to normalize
    """
    if isinstance(obj, Mapping):
        return obj

    if obj is None:
        return {}

    if isinstance(obj, str):
        try:
            new_obj = yaml.safe_load(obj)
        except Exception:
            obj = {"": obj}
        else:
            if isinstance(obj, Mapping):
                obj = new_obj
        return obj

    raise ValueError('Return docs are not the correct type')


def normalize_option_type_names(obj):
    """Normalize common mispellings of type names."""
    if obj == 'boolean':
        return 'bool'

    if obj in ('integer', 'number'):
        return 'int'

    if obj in ('string', 'strings'):
        return 'str'

    if obj == 'dictionary':
        return 'dict'

    if obj == 'lists':
        return 'list'

    return obj


class LocalConfig:
    """Settings we want on all of our models."""

    extra = p.Extra.forbid


class BaseModel(p.BaseModel):
    """BaseModel that has our preferred default config."""

    # pydantic Config classes are pure datastructures with no builtin special data
    # pyre-ignore[15]
    Config = LocalConfig


class DeprecationSchema(BaseModel):
    """Schema for Deprecation Fields."""

    removed_in: str = REQUIRED_VERSION_F
    why: str
    alternative: str = ''

    @p.root_validator(pre=True)
    def rename_version(cls, values):
        """Make deprecations at this level match the toplevel name."""
        version = values.get('version', _SENTINEL)
        if version is not _SENTINEL:
            if values.get('removed_in'):
                raise ValueError('Cannot specify `version` if `removed_in`'
                                 ' has been specified.')

            values['removed_in'] = version
            del values['version']

        return values

    @p.root_validator(pre=True)
    def merge_typo_names(cls, values):
        alternatives = values.get('alternatives', _SENTINEL)

        if alternatives is not _SENTINEL:
            if values.get('alternative'):
                raise ValueError('Cannot specify `alternatives` if `alternative`'
                                 ' has been specified.')

            values['alternative'] = alternatives
            del values['alternatives']

        return values


class OptionsSchema(BaseModel):
    description: t.List[str]
    aliases: t.List[str] = []
    choices: t.List[t.Union[str, None]] = []
    default: JSONValueT = None
    deprecated: DeprecationSchema = p.Field({})
    elements: str = OPTION_TYPE_F
    required: bool = False
    type: str = OPTION_TYPE_F
    version_added: str = 'historical'

    @p.validator('aliases', 'description', 'choices', pre=True)
    def list_from_scalars(cls, obj):
        return list_from_scalars(obj)

    @p.validator('type', 'elements', pre=True)
    def normalize_option_type(cls, obj):
        return normalize_option_type_names(obj)

    @p.root_validator(pre=True)
    def get_rid_of_name(cls, values):
        """
        Remove name from this schema.

        ``name`` is redundant with ``description``.  If we did need a shorter description here for
        some reason, we should call it ``short_description`` to match the other parts of the schema.
        """
        if 'name' in values:
            del values['name']
        return values

    @p.root_validator(pre=True)
    def merge_typo_names(cls, values):
        element_type = values.get('element_type', _SENTINEL)

        if element_type is not _SENTINEL:
            if values.get('elements'):
                raise ValueError('Cannot specify `element_type` if `elements` has been specified.')

            values['elements'] = element_type
            del values['element_type']

        element = values.get('element', _SENTINEL)

        if element is not _SENTINEL:
            if values.get('elements'):
                raise ValueError('Cannot specify `element` if `elements` has been specified.')

            values['elements'] = element
            del values['element']

        return values


class SeeAlsoModSchema(BaseModel):
    module: str
    description: str = ""


class SeeAlsoRefSchema(BaseModel):
    description: str
    ref: str


class SeeAlsoLinkSchema(BaseModel):
    description: str
    link: str
    name: str


class DocSchema(BaseModel):
    description: t.List[str]
    name: str
    short_description: str
    aliases: t.List[str] = []
    author: t.List[str] = []
    deprecated: DeprecationSchema = p.Field({})
    extends_documentation_fragment: t.List[str] = []
    filename: str = ''
    notes: t.List[str] = []
    requirements: t.List[str] = []
    seealso: t.List[t.Union[SeeAlsoModSchema, SeeAlsoRefSchema, SeeAlsoLinkSchema]] = []
    todo: t.List[str] = []
    version_added: str = 'historical'

    @p.validator('author', 'description', 'extends_documentation_fragment', 'notes',
                 'requirements', 'todo', pre=True)
    def list_from_scalars(cls, obj):
        return list_from_scalars(obj)

    @p.root_validator(pre=True)
    def remove_plugin_type(cls, values):
        """
        Remove the plugin_type field from the doc.

        The caller should already know the plugin_type.

        If we decide to keep this, we need to figure out how to fill it in for every plugin; right
        now it's only set (manually) on inventory and shell.
        """
        if 'plugin_type' in values:
            del values['plugin_type']
        return values

    @p.root_validator(pre=True)
    def merge_plugin_names(cls, values):
        """
        Normalize the field which plugin names are in.

        Some plugin types are putting their names into fields named after the type of plugin.
        This makes it harder for code to make use of the data as it has to know what the plugin
        type is and then look up that field.  Standardize on a ``name`` field instead.
        """
        names = {}
        for name_field in ('become', 'cache', 'callback', 'cliconf', 'connection',
                           'httpapi', 'inventory', 'lookup', 'module', 'netconf',
                           'strategy', 'vars'):
            plugin_name = values.get(name_field, _SENTINEL)
            if plugin_name is not _SENTINEL:
                names[name_field] = plugin_name

        if names:
            if len(names) > 1:
                raise ValueError('Specified {0} but only one can be'
                                 ' specified.'.format(names.keys()))
            if values.get('name'):
                raise ValueError('Cannot specify {0} if `name` has been'
                                 '  specified.'.format(names.keys()))

            # This seems to be the best way to get key and value from a one-element dict
            for name_field, name_field_value in names.items():
                values['name'] = name_field_value
                del values[name_field]

        return values

    @p.root_validator(pre=True)
    def merge_typo_names(cls, values):
        cb_type = values.get('callback_type', _SENTINEL)

        if cb_type is not _SENTINEL:
            if values.get('type'):
                raise ValueError('Cannot specify `callback_type` if `type` has been specified.')

            values['type'] = cb_type
            del values['callback_type']

        note = values.get('note', _SENTINEL)

        if note is not _SENTINEL:
            if values.get('notes'):
                raise ValueError('Cannot specify `note` if `notes` has been specified.')

            values['notes'] = note
            del values['note']

        authors = values.get('authors', _SENTINEL)

        if authors is not _SENTINEL:
            if values.get('author'):
                raise ValueError('Cannot specify `authors` if `author` has been specified.')

            values['author'] = authors
            del values['authors']

        return values



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Models for the antsibull config file."""

# Ignore Unitialized attribute errors because BaseModel works some magic
# to initialize the attributes when data is loaded into them.
# pyre-ignore-all-errors[13]


import typing as t

import pydantic as p
import twiggy.formats
import twiggy.outputs


LEVEL_CHOICES_F = p.Field(..., regex='^(CRITICAL|ERROR|WARNING|NOTICE|INFO|DEBUG|DISABLED)$')
VERSION_CHOICES_F = p.Field(..., regex=r'1\.0')


class FiltersSchema(p.BaseModel):
    filter: t.Union[str, t.Callable]
    args: t.Sequence[t.Any] = []
    kwargs: t.Mapping[str, t.Any] = {}


class EmitterSchema(p.BaseModel):
    level: str = LEVEL_CHOICES_F
    output_name: str
    filters: t.Optional[FiltersSchema] = None


class OutputSchema(p.BaseModel):
    output: t.Union[str, t.Callable]
    args: t.Sequence[t.Any] = []
    format: t.Union[str, t.Callable] = twiggy.formats.line_format
    kwargs: t.Mapping[str, t.Any] = {}


class ConfigSchema(p.BaseModel):
    emitters: t.Optional[t.Dict[str, EmitterSchema]] = {}
    incremental: bool = False
    outputs: t.Optional[t.Dict[str, OutputSchema]] = {}
    version: str = VERSION_CHOICES_F



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Functions to handle config files."""


import perky

from .logging import log
from .schemas.config import ConfigSchema


mlog = log.fields(mod=__name__)


def load_config(filename: str) -> ConfigSchema:
    """
    Parse a config file and return the data from it.

    :arg filename: The filename of the config file to parse.
    :returns: A ConfigSchema model containing the data.
    """
    raw_config_data = perky.load(filename)
    config_data = ConfigSchema.parse_obj(raw_config_data)
    return config_data



# This code is part of Ansible, but is an independent component.
# This particular file snippet, and this file snippet only, is BSD licensed.
# Modules you write using this snippet, which is embedded dynamically by Ansible
# still belong to the author of the module, and may assign their own license
# to the complete work.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# NB: a copy of this function exists in ../../modules/core/async_wrapper.py. Ensure any
# changes are propagated there.
def _filter_non_json_lines(data):
    '''
    Used to filter unrelated output around module JSON output, like messages from
    tcagetattr, or where dropbear spews MOTD on every single command (which is nuts).

    Filters leading lines before first line-starting occurrence of '{' or '[', and filter all
    trailing lines after matching close character (working from the bottom of output).
    '''
    warnings = []

    # Filter initial junk
    lines = data.splitlines()

    for start, line in enumerate(lines):
        line = line.strip()
        if line.startswith(u'{'):
            endchar = u'}'
            break
        elif line.startswith(u'['):
            endchar = u']'
            break
    else:
        raise ValueError('No start of json char found')

    # Filter trailing junk
    lines = lines[start:]

    for reverse_end_offset, line in enumerate(reversed(lines)):
        if line.strip().endswith(endchar):
            break
    else:
        raise ValueError('No end of json char found')

    if reverse_end_offset > 0:
        # Trailing junk is uncommon and can point to things the user might
        # want to change.  So print a warning if we find any
        trailing_junk = lines[len(lines) - reverse_end_offset:]
        for line in trailing_junk:
            if line.strip():
                warnings.append('Module invocation had junk after the JSON data: %s'
                                % '\n'.join(trailing_junk))
                break

    lines = lines[:(len(lines) - reverse_end_offset)]

    return ('\n'.join(lines), warnings)



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Entrypoint to the antsibull-build tool."""

import argparse
import os.path
import sys
from typing import List

import packaging.version as pypiver

from ..new_acd import new_acd_command
from ..build_collection import build_collection_command
from ..build_acd_commands import build_single_command, build_multiple_command


DEFAULT_FILE_BASE = 'acd'
DEFAULT_PIECES_FILE = f'{DEFAULT_FILE_BASE}.in'

ARGS_MAP = {'new-acd': new_acd_command,
            'build-single': build_single_command,
            'build-multiple': build_multiple_command,
            'build-collection': build_collection_command,
            }


class InvalidArgumentError(Exception):
    """A problem parsing or validating a command line argument."""


def _normalize_common_options(args: argparse.Namespace) -> None:
    if args.command is None:
        raise InvalidArgumentError('Please specify a subcommand to run')

    args.dest_dir = os.path.expanduser(os.path.expandvars(args.dest_dir))
    if not os.path.isdir(args.dest_dir):
        raise InvalidArgumentError(f'{args.dest_dir} must be an existing directory')


def _normalize_new_release_options(args: argparse.Namespace) -> None:
    if args.command != 'new-acd':
        return

    if args.pieces_file is None:
        args.pieces_file = os.path.join(args.dest_dir, DEFAULT_PIECES_FILE)

    if not os.path.isfile(args.pieces_file):
        raise InvalidArgumentError(f'The pieces file, {args.pieces_file}, must already'
                                   ' exist. It should contain one namespace.collection'
                                   ' per line')

    if args.build_file is None:
        basename = os.path.basename(os.path.splitext(args.pieces_file)[0])
        args.build_file = f'{basename}-{args.acd_version.major}.{args.acd_version.minor}.build'


def _normalize_release_build_options(args: argparse.Namespace) -> None:
    if args.command not in ('build-single', 'build-multiple'):
        return

    if args.build_file is None:
        args.build_file = (DEFAULT_FILE_BASE
                           + f'-{args.acd_version.major}.{args.acd_version.minor}.build')

    if not os.path.isfile(args.build_file):
        raise InvalidArgumentError(f'The build file, {args.build_file} must already exist.'
                                   ' It should contains one namespace.collection and range'
                                   ' of versions per line')

    if args.deps_file is None:
        major_minor = f'-{args.acd_version.major}.{args.acd_version.minor}'
        basename = os.path.basename(os.path.splitext(args.build_file)[0])
        if basename.endswith(major_minor):
            basename = basename[:-len(major_minor)]

        args.deps_file = f'{basename}-{args.acd_version}.deps'


def _normalize_collection_build_options(args: argparse.Namespace) -> None:
    if args.command != 'build-collection':
        return

    if args.deps_file is None:
        args.deps_file = DEFAULT_FILE_BASE + f'{args.acd_version}.deps'


def parse_args(program_name: str, args: List[str]) -> argparse.Namespace:
    """
    Parse and coerce the command line arguments.

    :arg program_name: The name of the program
    :arg args: A list of the command line arguments
    :returns: A :python:`argparse.Namespace`
    :raises InvalidArgumentError: Whenever there's something wrong with the arguments.
    """
    common_parser = argparse.ArgumentParser(add_help=False)
    common_parser.add_argument('acd_version', type=pypiver.Version,
                               help='The X.Y.Z version of ACD that this will be for')
    common_parser.add_argument('--dest-dir', default='.',
                               help='Directory to write the output to')

    build_parser = argparse.ArgumentParser(add_help=False)
    build_parser.add_argument('--build-file', default=None,
                              help='File containing the list of collections with version ranges.'
                              ' The default is to look for $DEFAULT_FILE_BASE-X.Y.build inside'
                              ' of --dest-dir')
    build_parser.add_argument('--deps-file', default=None,
                              help='File which will be written containing the list of collections'
                              ' at versions which were included in this version of ACD. The'
                              ' default is to place $BASENAME_OF_BUILD_FILE-X.Y.Z.deps into'
                              ' --dest-dir')

    parser = argparse.ArgumentParser(prog=program_name,
                                     description='Script to manage building ACD')
    subparsers = parser.add_subparsers(title='Subcommands', dest='command',
                                       help='for help use antsibull-build SUBCOMMANDS -h')

    new_parser = subparsers.add_parser('new-acd', parents=[common_parser],
                                       description='Generate a new build description from the'
                                       ' latest available versions of ansible-base and the'
                                       ' included collections')
    new_parser.add_argument('--pieces-file', default=None,
                            help='File containing a list of collections to include.  The'
                            f' default is to look for {DEFAULT_PIECES_FILE} inside of --dest-dir')
    new_parser.add_argument('--build-file', default=None,
                            help='File which will be written which contains the list'
                            ' of collections with version ranges.  The default is to'
                            ' place $BASENAME_OF_PIECES_FILE-X.Y.build into --dest-dir')

    build_single_parser = subparsers.add_parser('build-single',
                                                parents=[common_parser, build_parser],
                                                description='Build a single-file ACD')

    build_single_parser.add_argument('--debian', action='store_true',
                                     help='Include Debian/Ubuntu packaging files in'
                                     ' the resulting output directory')

    subparsers.add_parser('build-multiple',
                          parents=[common_parser, build_parser],
                          description='Build a multi-file ACD')

    collection_parser = subparsers.add_parser('build-collection',
                                              parents=[common_parser],
                                              description='Build a collection which will'
                                              ' install ACD')
    collection_parser.add_argument('--deps-file', default=None,
                                   help='File which contains the list of collections and'
                                   ' versions which were included in this version of ACD'
                                   f' The default is to look for {DEFAULT_FILE_BASE}-X.Y.Z.deps'
                                   ' inside of --dest-dir')

    args: argparse.Namespace = parser.parse_args(args)

    # Validation and coercion
    _normalize_common_options(args)
    _normalize_new_release_options(args)
    _normalize_release_build_options(args)
    _normalize_collection_build_options(args)

    return args


def run(args: List[str]) -> int:
    """
    Run the program.

    :arg args: A list of command line arguments.  Typically :python:`sys.argv`.
    :returns: A program return code.  0 for success, integers for any errors.  These are documented
        in :func:`main`.
    """
    program_name = os.path.basename(args[0])
    try:
        args: argparse.Namespace = parse_args(program_name, args[1:])
    except InvalidArgumentError as e:
        print(e)
        return 2

    return ARGS_MAP[args.command](args)


def main() -> int:
    """
    Entrypoint called from the script.

    console_scripts call functions which take no parameters.  However, it's hard to test a function
    which takes no parameters so this function lightly wraps :func:`run`, which actually does the
    heavy lifting.

    :returns: A program return code.

    Return codes:
        :0: Success
        :1: Unhandled error.  See the Traceback for more information.
        :2: There was a problem with the command line arguments
        :3: version in an input file does not match with the version specified on the command line
        :4: Needs to be run on a newer version of Python
    """
    if sys.version_info < (3, 8):
        print('Needs Python 3.8 or later')
        return 4

    return run(sys.argv)
